# -*- coding: utf-8 -*-
#
# File: testFaceted.py
#
# GNU General Public License (GPL)
#

from AccessControl import Unauthorized
from collective.contact.plonegroup.utils import get_organizations
from datetime import datetime
from eea.facetednavigation.interfaces import IFacetedLayout
from imio.helpers.content import get_vocab
from imio.helpers.content import get_vocab_values
from Products.Archetypes.event import ObjectEditedEvent
from Products.PloneMeeting.config import ITEM_NO_PREFERRED_MEETING_VALUE
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent


class testFaceted(PloneMeetingTestCase):
    '''Tests various aspects of faceted navigation.'''

    def test_pm_RedirectedToOwnPMFolderIfOnAnotherUserPMFolder(self):
        '''In case a user is sent to another user pmFolder, he is redirected
           to his own pmFolder even for Plone Managers, except if it is the Zope admin.'''
        cfgId = self.meetingConfig.getId()
        # get the pmCreator1 pmFolder
        self.changeUser('pmCreator1')
        self.request.RESPONSE.setHeader('location', '')
        creatorPMFolder = self.tool.getPloneMeetingFolder(cfgId)
        self.assertEqual(creatorPMFolder.getLayout(), 'facetednavigation_view')
        # check faceted_layout
        self.assertEqual(IFacetedLayout(creatorPMFolder).layout, 'faceted-table-items')
        creatorPMFolderUrl = creatorPMFolder.absolute_url()
        # access the pmFolder
        creatorPMFolder()
        # user was redirected to his pmFolder '/searches_items'
        self.assertTrue(self.request.RESPONSE.getStatus() == 302)
        self.assertTrue(
            self.request.RESPONSE.getHeader('location') == creatorPMFolderUrl + '/searches_items')

        # as another simple user or MeetingManager, Unauthorized as other pm folder is private
        self.changeUser('pmReviewer1')
        self.assertRaises(Unauthorized, creatorPMFolder.restrictedTraverse, '@@facetednavigation_view')
        self.changeUser('pmManager')
        self.assertRaises(Unauthorized, creatorPMFolder.restrictedTraverse, '@@facetednavigation_view')

        # as a Plone admin, the user is redirected to it's own pmFolder
        self.changeUser('siteadmin')
        self.request.RESPONSE.setHeader('location', '')
        siteadminPMFolder = self.tool.getPloneMeetingFolder(cfgId)
        siteadminPMFolderUrl = siteadminPMFolder.absolute_url()
        creatorPMFolder()
        self.assertEqual(
            self.request.RESPONSE.getHeader('location'), siteadminPMFolderUrl + '/searches_items')

        # if a user is using folder_contents, then he is not redirected
        self.request.RESPONSE.setHeader('location', '')
        self.request.RESPONSE.setStatus(200)
        creatorPMFolder.restrictedTraverse('folder_contents')()
        # user is not redirected
        self.assertFalse(self.request.RESPONSE.getHeader('location'))
        self.assertTrue(self.request.RESPONSE.getStatus() == 200)
        # make sure the different searches_... views are working
        # check some searches and faceted filters
        searches_items = creatorPMFolder.searches_items()
        self.assertTrue('Items to validate' in searches_items)
        self.assertTrue('section-preferred-meeting' in searches_items)
        self.assertTrue('section-to-send-to' in searches_items)
        searches_meetings = creatorPMFolder.searches_meetings()
        self.assertTrue('Items to validate' in searches_meetings)
        self.assertTrue("section-review-state" in searches_meetings)
        self.assertTrue("section-date" in searches_meetings)
        searches_decisions = creatorPMFolder.searches_decisions()
        self.assertTrue('Items to validate' in searches_decisions)
        self.assertTrue("section-review-state" in searches_decisions)
        self.assertTrue("section-date" in searches_decisions)

    def test_pm_RedirectToNextMeeting(self):
        """When selected, some user profiles will be redirected to the next meeting if it exists
           instead a dashboard displaying items (my items, ...)."""
        cfg = self.meetingConfig
        self.assertEqual(cfg.getRedirectToNextMeeting(), ())
        cfgId = cfg.getId()
        # get the pmCreator1 pmFolder
        self.changeUser('pmCreator1')
        self.request.RESPONSE.setHeader('location', '')
        creatorPMFolder = self.tool.getPloneMeetingFolder(cfgId)
        # access the pmFolder
        creatorPMFolder.searches_items()
        self.assertEqual(self.request.RESPONSE.getHeader('location'), '')
        # redirect app user to next meeting
        self.assertEqual(cfg.listRedirectToNextMeeting().keys(),
                         ['app_users',
                          'meeting_managers',
                          'powerobserver__powerobservers',
                          'powerobserver__restrictedpowerobservers'])
        cfg.setRedirectToNextMeeting(('app_users', ))
        # still searches_items as no meeting exist
        creatorPMFolder.searches_items()
        self.assertEqual(self.request.RESPONSE.getStatus(), 200)
        self.assertEqual(self.request.RESPONSE.getHeader('location'), '')
        self.changeUser('pmManager')
        meeting = self.create('Meeting')
        meeting_url = meeting.absolute_url()
        # freeze meeting so we are sure it is viewable (depends on WF)
        self.freezeMeeting(meeting)
        self.changeUser('pmCreator1')
        creatorPMFolder.searches_items()
        self.assertEqual(self.request.RESPONSE.getStatus(), 302)
        self.assertEqual(self.request.RESPONSE.getHeader('location'), meeting_url)

    def test_pm_RedirectToNextMeetingWhenMeetingNotViewable(self):
        """If meeting not viewable, user not redirected."""
        cfg = self.meetingConfig
        cfgId = cfg.getId()
        self._setPowerObserverStates(field_name='meeting_states',
                                     states=('closed', ))
        self.assertEqual(cfg.listRedirectToNextMeeting().keys(),
                         ['app_users',
                          'meeting_managers',
                          'powerobserver__powerobservers',
                          'powerobserver__restrictedpowerobservers'])
        cfg.setRedirectToNextMeeting(('powerobserver__powerobservers', ))
        self.changeUser('pmManager')
        meeting = self.create('Meeting')
        meeting_url = meeting.absolute_url()
        # not viewable for now
        self.changeUser('powerobserver1')
        pmFolder = self.tool.getPloneMeetingFolder(cfgId)
        # access the pmFolder
        pmFolder.searches_items()
        self.assertEqual(self.request.RESPONSE.getStatus(), 200)
        self.assertEqual(self.request.RESPONSE.getHeader('location'), '')
        # make meeting viewable
        self.closeMeeting(meeting, as_manager=True)
        # when no_redirect in request, user is not redirected
        self.assertEqual(self.request.get('no_redirect'), '1')
        pmFolder.searches_items()
        self.assertEqual(self.request.RESPONSE.getStatus(), 200)
        self.assertEqual(self.request.RESPONSE.getHeader('location'), '')
        # redirected
        self.request.set('no_redirect', 0)
        pmFolder.searches_items()
        self.assertEqual(self.request.RESPONSE.getStatus(), 302)
        self.assertEqual(self.request.RESPONSE.getHeader('location'), meeting_url)

    def test_pm_ItemCategoriesVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.categoriesvocabulary"
           vocabulary, especially because it is cached.'''
        self.changeUser('siteadmin')
        pmFolder = self.getMeetingFolder()
        cfg = self.meetingConfig
        self._enableField('category')
        vocab = get_vocab(cfg,
                          "Products.PloneMeeting.vocabularies.categoriesvocabulary",
                          only_factory=True)
        # once get, it is cached
        terms = vocab(pmFolder)
        # every existing categories are shown, no matter it is disabled
        nbOfCategories = len(cfg.getCategories(onlySelectable=False))
        self.assertEqual(len(terms), nbOfCategories)
        # here we make sure it is cached by changing a category title
        # manually without using the processForm way
        dev = cfg.categories.development
        dev.title = u'New title'
        terms = vocab(pmFolder)
        self.assertNotEquals(terms.by_token['development'].title,
                             cfg.categories.development.title)
        # right correctly edit the category, the vocabulary is invalidated
        notify(ObjectModifiedEvent(dev))
        terms = vocab(pmFolder)
        self.assertEqual(terms.by_token['development'].title,
                         cfg.categories.development.title)

        # if we add/remove a category, then the cache is cleaned too
        # add a category
        newCat = self.create('meetingcategory',
                             id='new-category',
                             title='New category')
        # cache was cleaned, the new value is available
        terms = vocab(pmFolder)
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Events', u'New category', u'New title', u'Research topics'])

        # disable a category
        self._disableObj(newCat)
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Events', u'New title', u'Research topics', u'New category (Inactive)'])
        # term.value is the category id
        self.assertEqual(
            [term.value for term in vocab(pmFolder)],
            [u'events', u'development', u'research', u'new-category'])

        # remove a category
        self.portal.restrictedTraverse('@@delete_givenuid')(newCat.UID())
        # cache was cleaned
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Events', u'New title', u'Research topics'])

    def test_pm_ItemClassifiersVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.classifiersvocabulary"
           vocabulary, especially because it is cached. It relies on the categoriesvocabulary.'''
        self.changeUser('siteadmin')
        pmFolder = self.getMeetingFolder()
        cfg = self.meetingConfig
        self._enableField('category')
        vocab = get_vocab(cfg,
                          "Products.PloneMeeting.vocabularies.classifiersvocabulary",
                          only_factory=True)
        # once get, it is cached
        terms = vocab(pmFolder)
        # every existing categories are shown, no matter it is disabled
        nbOfCategories = len(cfg.getCategories(catType='classifiers', onlySelectable=False))
        self.assertEqual(len(terms), nbOfCategories)
        # here we make sure it is cached by changing a category title
        # manually without using the processForm way
        classifier1 = cfg.classifiers.classifier1
        classifier1.title = u'New title'
        terms = vocab(pmFolder)
        classifier1_id = classifier1.getId()
        self.assertNotEquals(terms.by_token[classifier1_id].title,
                             cfg.categories.development.title)
        # right correctly edit the category, the vocabulary is invalidated
        notify(ObjectModifiedEvent(classifier1))
        terms = vocab(pmFolder)
        self.assertEqual(terms.by_token[classifier1_id].title,
                         cfg.classifiers.classifier1.title)

        # if we add/remove a category, then the cache is cleaned too
        # add a classifier
        newClassifier = self.create('meetingcategory',
                                    id='newclassifier',
                                    title='New classifier',
                                    is_classifier=True)
        # cache was cleaned, the new value is available
        terms = vocab(pmFolder)
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Classifier 2', u'Classifier 3', u'New classifier', u'New title'])

        # disable a classifier
        self._disableObj(newClassifier)
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Classifier 2', u'Classifier 3', u'New title', u'New classifier (Inactive)'])
        # term.value is the classifier id
        self.assertEqual(
            [term.value for term in vocab(pmFolder)],
            ['classifier2', 'classifier3', 'classifier1', 'newclassifier'])

        # remove a classifier
        self.portal.restrictedTraverse('@@delete_givenuid')(newClassifier.UID())
        # cache was cleaned
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Classifier 2', u'Classifier 3', u'New title'])

    def test_pm_ItemCategoriesVocabularyMCAware(self):
        '''Test that "Products.PloneMeeting.vocabularies.categoriesvocabulary"
           vocabulary, is MeetingConfig aware, especially because it is cached.'''
        self.changeUser('siteadmin')
        pmFolder = self.getMeetingFolder()
        cfg = self.meetingConfig
        self._enableField('category')
        vocab = get_vocab(cfg,
                          "Products.PloneMeeting.vocabularies.categoriesvocabulary",
                          only_factory=True)
        terms_cfg1 = [term.token for term in vocab(pmFolder)]
        # now in cfg2
        cfg2 = self.meetingConfig2
        self.setMeetingConfig(cfg2.getId())
        self._enableField('category', cfg=cfg2)
        pmFolder = self.getMeetingFolder()
        terms_cfg2 = [term.token for term in vocab(pmFolder)]
        self.assertNotEqual(terms_cfg1, terms_cfg2)

    def test_pm_MeetingDatesVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.meetingdatesvocabulary"
           vocabulary, especially because it is cached.'''
        self._removeConfigObjectsFor(self.meetingConfig)
        self.changeUser('pmManager')
        pmFolder = self.getMeetingFolder()
        # create a meeting
        meeting = self.create('Meeting', date=datetime(2015, 5, 5))
        meetingUID = meeting.UID()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.meetingdatesvocabulary",
                          only_factory=True)
        # once get, it is cached
        vocab(pmFolder)
        self.assertEqual(
            [term.token for term in vocab(pmFolder)._terms],
            [ITEM_NO_PREFERRED_MEETING_VALUE, meetingUID])

        # if we add/remove/edit a meeting, then the cache is cleaned
        # add a meeting
        meeting2 = self.create('Meeting', date=datetime(2015, 6, 6))
        meeting2UID = meeting2.UID()
        # cache was cleaned
        self.assertEqual(
            [term.token for term in vocab(pmFolder)._terms],
            [ITEM_NO_PREFERRED_MEETING_VALUE, meeting2UID, meetingUID])
        # edit a meeting
        self.assertEqual(vocab(pmFolder).by_token[meetingUID].title, u'05/05/2015')
        meeting.date = datetime(2015, 7, 7)
        meeting._update_after_edit()
        # cache was cleaned
        self.assertEqual(vocab(pmFolder).by_token[meetingUID].title, u'07/07/2015')

        # remove a meeting
        self.portal.restrictedTraverse('@@delete_givenuid')(meeting.UID())
        # cache was cleaned
        self.assertEqual(
            [term.token for term in vocab(pmFolder)._terms],
            [ITEM_NO_PREFERRED_MEETING_VALUE, meeting2UID])

    def test_pm_MeetingDatesVocabularyMCAware(self):
        '''Test that "Products.PloneMeeting.vocabularies.meetingdatesvocabulary"
           vocabulary, is MeetingConfig aware, especially because it is cached.'''
        cfg = self.meetingConfig
        cfg2 = self.meetingConfig2
        self.changeUser('pmManager')
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.meetingdatesvocabulary",
                          only_factory=True)
        # create Meetings in cfg1
        self.create('Meeting')
        self.create('Meeting', date=datetime(2015, 5, 6))
        # create Meetings in cfg2
        self.setMeetingConfig(cfg2.getId())
        self.create('Meeting')
        self.create('Meeting', date=datetime(2016, 5, 6))

        self.setMeetingConfig(cfg.getId())
        pmFolder = self.getMeetingFolder()
        terms_cfg1 = [term.token for term in vocab(pmFolder)]
        self.setMeetingConfig(cfg2.getId())
        pmFolder = self.getMeetingFolder()
        terms_cfg2 = [term.token for term in vocab(pmFolder)]
        self.assertNotEqual(terms_cfg1, terms_cfg2)

    def _orgs_to_exclude_from_filter(self):
        return ()

    def test_pm_ProposingGroupsVocabularies(self):
        '''Test proposingGroup related cached vocabularies.'''
        self.changeUser('siteadmin')
        # check test profile values
        self.assertGreater(len(self.proposing_groups), 2)
        # greater than proposing_groups + "My organization"
        self.assertGreater(len(self.all_org), len(self.proposing_groups) + 1)
        # not all proposing groups are activated
        self.assertLess(len(self.active_proposing_groups), len(self.proposing_groups))

        pmFolder = self.getMeetingFolder()
        vocab1 = get_vocab(
            pmFolder,
            "Products.PloneMeeting.vocabularies.proposinggroupsvocabulary",
            only_factory=True)
        vocab2 = get_vocab(
            pmFolder,
            "Products.PloneMeeting.vocabularies.everyorganizationsacronymsvocabulary",
            only_factory=True)
        vocab3 = get_vocab(
            pmFolder,
            "Products.PloneMeeting.vocabularies.proposinggroupsforfacetedfiltervocabulary",
            only_factory=True)
        vocab4 = get_vocab(
            pmFolder,
            "Products.PloneMeeting.vocabularies.associatedgroupsvocabulary",
            only_factory=True)
        # once get, it is cached
        self.assertEqual(len(vocab1(pmFolder)), len(self.proposing_groups))
        # contains My organization and external organizations
        self.assertEqual(len(vocab2(pmFolder)), len(self.all_org))
        self.assertEqual(len(vocab3(pmFolder)), len(self.proposing_groups))
        # when nothing in config, just displays the orgs selected in plonegroup
        self.assertEqual(len(vocab4(pmFolder)), len(self.active_proposing_groups))

        # if we add/remove/edit an organozation, then the cache is cleaned
        # add an organization
        new_org = self.create('organization', title='NewOrg', acronym='N.O.')
        new_org_uid = new_org.UID()
        self._select_organization(new_org_uid)
        # cache was cleaned
        self.assertEqual(len(vocab1(pmFolder)), len(self.proposing_groups) + 1)
        self.assertEqual(len(vocab2(pmFolder)), len(self.all_org) + 1)
        self.assertEqual(len(vocab3(pmFolder)), len(self.proposing_groups) + 1)
        self.assertEqual(len(vocab4(pmFolder)), len(self.active_proposing_groups) + 1)

        # edit a group
        self.assertEqual(vocab1(pmFolder).by_token[new_org_uid].title, new_org.Title())
        self.assertEqual(vocab2(pmFolder).by_token[new_org_uid].title, new_org.acronym)
        self.assertEqual(vocab3(pmFolder).by_token[new_org_uid].title, new_org.Title())
        self.assertEqual(vocab4(pmFolder).by_token[new_org_uid].title, new_org.Title())
        new_org.title = u'Modified title'
        new_org.acronym = u'Modified acronym'
        notify(ObjectModifiedEvent(new_org))
        # cache was cleaned
        self.assertEqual(vocab1(pmFolder).by_token[new_org_uid].title, new_org.Title())
        self.assertEqual(vocab2(pmFolder).by_token[new_org_uid].title, new_org.acronym)
        self.assertEqual(vocab3(pmFolder).by_token[new_org_uid].title, new_org.Title())
        self.assertEqual(vocab4(pmFolder).by_token[new_org_uid].title, new_org.Title())

        # remove an organization (unselect it first)
        self._select_organization(new_org_uid, remove=True)
        self.portal.restrictedTraverse('@@delete_givenuid')(new_org_uid)
        # cache was cleaned
        # once get, it is cached
        self.assertEqual(len(vocab1(pmFolder)), len(self.proposing_groups))
        self.assertEqual(len(vocab2(pmFolder)), len(self.all_org))
        self.assertEqual(len(vocab3(pmFolder)), len(self.proposing_groups))
        self.assertEqual(len(vocab4(pmFolder)), len(self.active_proposing_groups))
        # activate "End users"
        proposing_groups_terms_titles = sorted([group.title + (not group.active and u' (Inactive)' or u'')
                                                for group in self.proposing_groups])
        self.assertListEqual(
            sorted([term.title for term in vocab1(pmFolder)]), proposing_groups_terms_titles)
        self.assertListEqual(sorted([term.title for term in vocab2(pmFolder)]),
                             sorted([unicode(org.acronym) for org in self.all_org]))
        self.assertListEqual(
            sorted([term.title for term in vocab3(pmFolder)]), proposing_groups_terms_titles)

        associeted_groups = sorted([group.title for group in self.proposing_groups if group.active])
        self.assertListEqual([term.title for term in vocab4(pmFolder)], associeted_groups)
        self._select_organization(self.endUsers_uid)

        proposing_groups_terms_titles = sorted([group.title for group in self.proposing_groups])
        self.assertListEqual(
            sorted([term.title for term in vocab1(pmFolder)]), proposing_groups_terms_titles)
        self.assertListEqual(
            sorted([term.title for term in vocab2(pmFolder)]), sorted([unicode(org.acronym) for org in self.all_org]))
        self.assertListEqual(
            sorted([term.title for term in vocab3(pmFolder)]), proposing_groups_terms_titles)

        associeted_groups = sorted([group.title for group in self.proposing_groups])
        self.assertListEqual([term.title for term in vocab4(pmFolder)], associeted_groups)

    def test_pm_ProposingGroupsForFacetedVocabulary(self):
        '''Test that vocabulary "Products.PloneMeeting.vocabularies.proposinggroupsforfacetedfiltervocabulary"
           relies on MeetingConfig.groupsHiddenInDashboardFilter.'''
        cfg = self.meetingConfig
        self.changeUser('siteadmin')
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.proposinggroupsforfacetedfiltervocabulary",
                          only_factory=True)
        # by default when MeetingConfig.groupsHiddenInDashboardFilter is empty, every group are returned
        self.assertEqual(cfg.getGroupsHiddenInDashboardFilter(), ())
        # remove extra organizations from profiles
        cfg.setGroupsHiddenInDashboardFilter(self._orgs_to_exclude_from_filter())
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Developers', u'Vendors', u'End users (Inactive)'])
        # now define values in MeetingConfig.groupsHiddenInDashboardFilter
        cfg.setGroupsHiddenInDashboardFilter((self.vendors_uid, ) + self._orgs_to_exclude_from_filter())
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Developers', u'End users (Inactive)'])

        # select "End users"
        self._select_organization(self.endUsers_uid)
        self.assertEqual(
            [term.title for term in vocab(pmFolder)],
            [u'Developers', u'End users'])

    def test_pm_GroupsInChargeVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.groupsinchargevocabulary"
           vocabulary, especially because it is cached.'''
        cfg = self.meetingConfig
        self._enableField('category')
        self._enableField('classifier')
        self.changeUser('siteadmin')
        org1 = self.create('organization', id='org1', title='Org 1', acronym='Org1')
        org1_uid = org1.UID()
        org2 = self.create('organization', id='org2', title='Org 2', acronym='Org2')
        org2_uid = org2.UID()
        org3 = self.create('organization', id='org3', title='Org 3', acronym='Org3')
        org3_uid = org3.UID()
        vocab = get_vocab(org1,
                          "Products.PloneMeeting.vocabularies.groupsinchargevocabulary",
                          only_factory=True)

        # for now, no group in charge
        meetingFolder = self.getMeetingFolder()
        self.assertEqual(len(vocab(meetingFolder)), 0)
        # define some group in charge, vocabulary is invalidated when an org is modified
        self.vendors.groups_in_charge = (org1_uid,)
        notify(ObjectModifiedEvent(self.vendors))
        self.developers.groups_in_charge = (org2_uid,)
        notify(ObjectModifiedEvent(self.developers))
        self.assertEqual([term.title for term in vocab(meetingFolder)], ['Org 1', 'Org 2'])

        # create an new org with a groupInCharge directly
        org4 = self.create('organization', id='org4', title='Org 4',
                           acronym='Org4', groups_in_charge=(org3_uid,))
        org4_uid = org4.UID()
        self._select_organization(org4_uid)
        self.assertEqual([term.title for term in vocab(meetingFolder)], ['Org 1', 'Org 2', 'Org 3'])

        # change a group in charge
        self.vendors.groups_in_charge = (org4_uid,)
        notify(ObjectModifiedEvent(self.vendors))
        self.assertEqual([term.title for term in vocab(meetingFolder)], ['Org 2', 'Org 3', 'Org 4'])

        # unselect a group in charge
        self.vendors.groups_in_charge = ()
        notify(ObjectModifiedEvent(self.vendors))
        self.assertEqual([term.title for term in vocab(meetingFolder)], ['Org 2', 'Org 3'])

        # category, creating an organization invalidates the vocabulary cache
        org5 = self.create('organization', id='org5', title='Org 5', acronym='Org5')
        org5_uid = org5.UID()
        # an already used
        cfg.categories.development.groups_in_charge = [org3_uid]
        # already existing no more used
        cfg.categories.research.groups_in_charge = [org4_uid]
        # new
        cfg.categories.events.groups_in_charge = [org5_uid]
        self.assertEqual(
            [term.title for term in vocab(meetingFolder)],
            ['Org 2', 'Org 3', 'Org 4', 'Org 5'])

        # classifier, creating an organization invalidates the vocabulary cache
        org6 = self.create('organization', id='org6', title='Org 6', acronym='Org6')
        org6_uid = org6.UID()
        # already existing
        cfg.classifiers.classifier1.groups_in_charge = [org5_uid]
        # new
        cfg.classifiers.classifier2.groups_in_charge = [org6_uid]
        self.assertEqual(
            [term.title for term in vocab(meetingFolder)],
            ['Org 2', 'Org 3', 'Org 4', 'Org 5', 'Org 6'])

    def test_pm_CreatorsVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.creatorsvocabulary"
           vocabulary, especially because it is cached.'''
        self.changeUser('pmCreator1')
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.creatorsvocabulary",
                          only_factory=True)
        # once get, it is cached
        self.assertEqual(len(vocab(pmFolder)), 3)

        # if a new pmFolder is created, then the cache is cleaned
        # get pmFolder for user 'pmManager'
        self.changeUser('pmManager')
        pmFolder = self.getMeetingFolder()
        # cache was cleaned
        self.assertEqual(len(vocab(pmFolder)), 4)

    def test_pm_CreatorsForFacetedVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.creatorsforfacetedvocabulary"
                   vocabulary, especially because it is cached.'''
        cfg = self.meetingConfig
        self.changeUser('pmCreator1')
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(
            pmFolder,
            "Products.PloneMeeting.vocabularies.creatorsforfacetedfiltervocabulary",
            only_factory=True)
        # once get, it is cached
        self.assertEqual(len(vocab(pmFolder)), 3)

        # if a new pmFolder is created, then the cache is cleaned
        # get pmFolder for user 'pmManager'
        self.changeUser('pmManager')
        pmFolder = self.getMeetingFolder()
        # cache was cleaned
        self.assertEqual(len(vocab(pmFolder)), 4)

        cfg.setUsersHiddenInDashboardFilter(('pmCreator1',))
        notify(ObjectEditedEvent(cfg))
        # cache was cleaned and pmCreator is not in the list anymore
        self.assertEqual(len(vocab(pmFolder)), 3)

        cfg.setUsersHiddenInDashboardFilter(())
        notify(ObjectEditedEvent(cfg))
        # cache was cleaned and pmCreator is back in the list
        self.assertEqual(len(vocab(pmFolder)), 4)

    def test_pm_AskedAdvicesVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.askedadvicesvocabulary"
           vocabulary, especially because it is cached.'''
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        customAdvisers = [{'row_id': 'unique_id_000',
                           'org': self.developers_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '2',
                           'delay_label': ''},
                          {'row_id': 'unique_id_123',
                           'org': self.vendors_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '5',
                           'delay_label': ''},
                          {'row_id': 'unique_id_456',
                           'org': self.vendors_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '10',
                           'delay_label': ''},
                          {'row_id': 'unique_id_789',
                           'org': self.vendors_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '20',
                           'delay_label': ''}]
        cfg.setCustomAdvisers(customAdvisers)
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.askedadvicesvocabulary",
                          only_factory=True)
        # we have 4 delay-aware advisers and 2 adviser groups selectable as optional
        delayAdvisers = [adviser for adviser in cfg.getCustomAdvisers() if adviser['delay']]
        self.assertEqual(len(delayAdvisers), 4)
        self.assertEqual(len(get_organizations(not_empty_suffix='advisers')), 2)
        # once get, it is cached, it includes customAdvisers and MeetingConfig.selectableAdvisers
        self.assertEqual(len(vocab(pmFolder)), 6)

        # if we select a new organization, then the cache is cleaned
        # add an organization
        new_org = self.create('organization', title='New organization', acronym='N.G.')
        new_org_uid = new_org.UID()
        cfg.setSelectableAdvisers(cfg.getSelectableAdvisers() + (new_org_uid, ))
        notify(ObjectEditedEvent(cfg))
        # cache was cleaned
        self.assertEqual(len(vocab(pmFolder)), 7)
        # edit an organization
        new_org_term_id = 'real_org_uid__{0}'.format(new_org_uid)
        self.assertEqual(vocab(pmFolder).by_token[new_org_term_id].title, 'New organization (Inactive)')
        new_org.title = u'Modified title'
        notify(ObjectModifiedEvent(new_org))
        # cache was cleaned
        self.assertEqual(vocab(pmFolder).by_token[new_org_term_id].title, 'Modified title (Inactive)')
        # select the organization, cache is cleaned
        self._select_organization(new_org_uid)
        self.assertEqual(vocab(pmFolder).by_token[new_org_term_id].title, 'Modified title')
        # remove an organization
        # first need to unselect it
        self._select_organization(new_org_uid, remove=True)
        self.portal.restrictedTraverse('@@delete_givenuid')(new_org_uid)
        # cache was cleaned
        self.assertEqual(len(vocab(pmFolder)), 6)

        # if we add/remove/edit a customAdviser, then the cache is cleaned
        # add a customAdviser
        customAdvisers.append({'row_id': 'unique_id_999',
                               'org': self.vendors_uid,
                               'gives_auto_advice_on': '',
                               'for_item_created_from': '2012/01/01',
                               'delay': '11',
                               'delay_label': 'New delay'})
        cfg.setCustomAdvisers(customAdvisers)
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(len(vocab(pmFolder)), 7)
        self.assertTrue('delay_row_id__unique_id_999' in vocab(pmFolder).by_token)
        # delay is displayed in customAdviser title
        self.assertTrue('11 day(s)' in vocab(pmFolder).by_token['delay_row_id__unique_id_999'].title)
        # edit a customAdviser
        customAdvisers[-1]['delay'] = '12'
        cfg.setCustomAdvisers(customAdvisers)
        notify(ObjectEditedEvent(cfg))
        self.assertTrue('12 day(s)' in vocab(pmFolder).by_token['delay_row_id__unique_id_999'].title)
        # remove a customAdviser
        customAdvisers = customAdvisers[:-1]
        cfg.setCustomAdvisers(customAdvisers)
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(len(vocab(pmFolder)), 6)
        # power advisers are taken into account by the vocabulary
        cfg.setPowerAdvisersGroups([self.endUsers_uid])
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(len(vocab(pmFolder)), 7)
        # inactive term, displayed in term title
        # make a not really expired term, a 'for_item_created_until' in the future
        self.assertEqual(customAdvisers[-1]['row_id'], 'unique_id_789')
        customAdvisers[-1]['for_item_created_until'] = '2099/01/01'
        cfg.setCustomAdvisers(customAdvisers)
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(vocab(pmFolder).by_token['delay_row_id__unique_id_789'].title,
                         u'Vendors - 20 day(s)')
        customAdvisers[-1]['for_item_created_until'] = '2009/01/01'
        cfg.setCustomAdvisers(customAdvisers)
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(vocab(pmFolder).by_token['delay_row_id__unique_id_789'].title,
                         u'Vendors - 20 day(s) (Inactive)')

    def test_pm_AskedAdvicesVocabularyWithWrongContext(self):
        '''Test the "Products.PloneMeeting.vocabularies.askedadvicesvocabulary"
           vocabulary, when receiving wrong context, may occur during site install or
           DashboardCollection edit.'''
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        meeting = self.create('Meeting')
        cfg = self.meetingConfig
        vocab = get_vocab(cfg,
                          "Products.PloneMeeting.vocabularies.askedadvicesvocabulary",
                          only_factory=True)
        # working context
        pmFolder = self.getMeetingFolder()
        self.assertEqual(len(vocab(pmFolder)), 2)
        self.assertEqual(len(vocab(cfg)), 2)
        self.assertEqual(len(vocab(item)), 2)
        self.assertEqual(len(vocab(meeting)), 2)
        # do not fail on weird context
        self.assertEqual(len(vocab(self.portal)), 0)
        self.assertEqual(len(vocab(self.portal.contacts)), 0)
        self.assertEqual(len(vocab(self.app)), 0)
        self.request['PUBLISHED'] = None
        self.assertEqual(len(vocab(None)), 0)

    def test_pm_AskedAdvicesVocabularyMCAware(self):
        '''Test the "Products.PloneMeeting.vocabularies.askedadvicesvocabulary"
           vocabulary, is MeetingConfig aware, especially because it is cached.'''
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        cfg2 = self.meetingConfig2
        customAdvisers = [{'row_id': 'unique_id_000',
                           'org': self.developers_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '2',
                           'delay_label': ''},
                          {'row_id': 'unique_id_123',
                           'org': self.vendors_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '5',
                           'delay_label': ''},
                          {'row_id': 'unique_id_456',
                           'org': self.vendors_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '10',
                           'delay_label': ''},
                          {'row_id': 'unique_id_789',
                           'org': self.vendors_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '20',
                           'delay_label': ''}]
        cfg.setCustomAdvisers(customAdvisers)
        customAdvisers = [{'row_id': 'unique_id_999',
                           'org': self.developers_uid,
                           'gives_auto_advice_on': '',
                           'for_item_created_from': '2012/01/01',
                           'delay': '20',
                           'delay_label': ''}]
        cfg2.setCustomAdvisers(customAdvisers)
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.askedadvicesvocabulary",
                          only_factory=True)
        terms_cfg1 = [term.token for term in vocab(pmFolder)]
        self.setMeetingConfig(cfg2.getId())
        pmFolder = self.getMeetingFolder()
        terms_cfg2 = [term.token for term in vocab(pmFolder)]
        self.assertNotEqual(terms_cfg1, terms_cfg2)

    def test_pm_AdviceTypesVocabulary(self):
        '''Test the "Products.PloneMeeting.vocabularies.advicetypesvocabulary"
           vocabulary, especially because it is cached.'''
        cfg = self.meetingConfig
        self.changeUser('siteadmin')
        cfg.setUsedAdviceTypes(('positive', 'negative'))
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.advicetypesvocabulary",
                          only_factory=True)
        # once get, it is cached
        self.assertEqual(sorted([term.token for term in vocab(pmFolder)]),
                         ['asked_again',
                          'considered_not_given_hidden_during_redaction',
                          'hidden_during_redaction',
                          'negative',
                          'not_given',
                          'positive'])

        # change the MeetingConfig.usedAdvicesTypes
        cfg.setUsedAdviceTypes(('positive', ))
        # cached
        self.assertEqual(sorted([term.token for term in vocab(pmFolder)]),
                         ['asked_again',
                          'considered_not_given_hidden_during_redaction',
                          'hidden_during_redaction',
                          'negative',
                          'not_given',
                          'positive'])
        notify(ObjectEditedEvent(cfg))
        # cache invalidated
        self.assertEqual(sorted([term.token for term in vocab(pmFolder)]),
                         ['asked_again',
                          'considered_not_given_hidden_during_redaction',
                          'hidden_during_redaction',
                          'not_given',
                          'positive'])
        # ToolPloneMeeting.advisersConfig.advice_types
        self.tool.setAdvisersConfig(
            ({'advice_types': ['positive_with_remarks'],
              'base_wf': 'meetingadvice_workflow',
              'default_advice_type': 'positive_with_remarks',
              'org_uids': [self.vendors_uid],
              'portal_type': 'meetingadvice',
              'show_advice_on_final_wf_transition': '1',
              'wf_adaptations': []}, ))
        self.tool.at_post_edit_script()
        notify(ObjectEditedEvent(cfg))
        self.assertEqual(sorted([term.token for term in vocab(pmFolder)]),
                         ['asked_again',
                          'considered_not_given_hidden_during_redaction',
                          'hidden_during_redaction',
                          'not_given',
                          'positive',
                          'positive_with_remarks'])

    def test_pm_PMConditionAwareCollectionVocabulary(self):
        """Test the 'conditionawarecollectionvocabulary'
           essentially because it is cached. """
        # test with 2 members having same groups
        # as pmCreator1
        self.changeUser('pmCreator1')
        creator1_groups = self.member.getGroups()
        pmFolder = self.getMeetingFolder()
        searches = self.meetingConfig.searches
        searchAllItems = searches.searches_items.searchallitems
        vocab_name = "Products.PloneMeeting.vocabularies.conditionawarecollectionvocabulary"
        vocab = get_vocab(pmFolder, vocab_name, only_factory=True)
        self.assertTrue("/pmCreator1/" in vocab(searchAllItems, pmFolder)._terms[0].title[1])
        # as pmCreator1b
        self.changeUser('pmCreator1b')
        creator1_groups = self.member.getGroups()
        creator1b_groups = self.member.getGroups()
        self.assertEqual(creator1_groups, creator1b_groups)
        self.assertTrue("/pmCreator1b/" in vocab(searchAllItems, pmFolder)._terms[0].title[1])
        # invalidated if a collection is edited
        self.assertEqual(vocab(searchAllItems, pmFolder)._terms[0].title[0], u'My items')
        searchMyItems = searches.searches_items.searchmyitems
        searchMyItems.setTitle(u'My items edited')
        notify(ObjectModifiedEvent(searchMyItems))
        self.assertEqual(vocab(searchAllItems, pmFolder)._terms[0].title[0], u'My items edited')
        # invalidated when user groups changed
        # make pmCreator1b no more a creator
        vocab_values = get_vocab_values(searchAllItems, vocab_name, **{'real_context': pmFolder})
        self._removePrincipalFromGroups(self.member.id, [self.developers_creators])
        no_group_vocab_values = get_vocab_values(searchAllItems, vocab_name, **{'real_context': pmFolder})
        self.assertNotEqual(vocab_values, no_group_vocab_values)
        self._addPrincipalToGroup(self.member.id, self.developers_observers)
        observer_vocab_values = get_vocab_values(searchAllItems, vocab_name, **{'real_context': pmFolder})
        self.assertNotEqual(no_group_vocab_values, observer_vocab_values)

    def test_pm_AdviceTypesVocabularyMCAware(self):
        '''Test the "Products.PloneMeeting.vocabularies.advicetypesvocabulary"
           vocabulary, is MeetingConfig aware, especially because it is cached.'''
        cfg = self.meetingConfig
        cfg2 = self.meetingConfig2
        self.changeUser('siteadmin')
        cfg.setUsedAdviceTypes(('positive', 'negative'))
        cfg2.setUsedAdviceTypes(('positive', ))
        vocab = get_vocab(cfg,
                          "Products.PloneMeeting.vocabularies.advicetypesvocabulary",
                          only_factory=True)
        pmFolder = self.getMeetingFolder()
        terms_cfg1 = [term.token for term in vocab(pmFolder)]
        self.setMeetingConfig(cfg2.getId())
        pmFolder = self.getMeetingFolder()
        terms_cfg2 = [term.token for term in vocab(pmFolder)]
        self.assertNotEqual(terms_cfg1, terms_cfg2)

    def test_pm_RedirectedToDefaultSearchPMFolderOnlyIfNecessary(self):
        """This test portlet_plonemeeting.widget_render where we manipulate the redirection,
           returned by the collection widget to the default collection as collections are in the configuration
           and we want the user to be redirected in his meeting folder."""
        self.changeUser('pmCreator1')
        pmFolder = self.getMeetingFolder()
        item = self.create('MeetingItem')

        # call to pmFolder redirects to searches_items
        self.assertEqual(self.request.RESPONSE.status, 200)
        pmFolder()
        self.assertEqual(self.request.RESPONSE.status, 302)
        self.assertEqual(
            self.request.RESPONSE.getHeader('location'), pmFolder.absolute_url() + '/searches_items')

        # not redirected if on an item
        self.request.RESPONSE.status = 200
        item()
        self.assertEqual(self.request.RESPONSE.status, 200)

        # not redirected if on an meeting
        self.changeUser('pmManager')
        pmFolder = self.getMeetingFolder()
        meeting = self.create('Meeting')
        view = meeting.restrictedTraverse('@@meeting_view')
        view()
        self.assertEqual(self.request.RESPONSE.status, 200)

    def test_pm_DisabledCollectionsAreIgnored(self):
        """If a DashboardCollection is disabled in the MeetingConfig,
           it is not displayed in the vocabulary."""
        searches = self.meetingConfig.searches
        searchAllItems = searches.searches_items.searchallitems
        searchAllItems_path = searchAllItems.absolute_url_path()
        self.changeUser('pmCreator1')
        pmFolder = self.getMeetingFolder()
        vocab = get_vocab(pmFolder,
                          "Products.PloneMeeting.vocabularies.conditionawarecollectionvocabulary",
                          only_factory=True)
        self.assertTrue(searchAllItems_path in vocab(searches, real_context=pmFolder))
        # disable it then test again
        self.changeUser('siteadmin')
        self._disableObj(searchAllItems, notify_event=True)
        self.changeUser('pmCreator1')
        self.assertFalse(searchAllItems_path in vocab(searches, real_context=pmFolder))

    def test_pm_SearchableTextQueryOnSpecialCharacters(self):
        """Make sure it is possible to query on "décision" and "decision"."""
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem', title="SpecialDécision")
        self.assertEqual(len(self.catalog(SearchableText="SpecialDécision")), 1)
        self.assertEqual(len(self.catalog(SearchableText="specialdécision")), 1)
        self.assertEqual(len(self.catalog(SearchableText="SpecialDecision")), 1)
        self.assertEqual(len(self.catalog(SearchableText="specialdecision")), 1)
        item.setTitle("SpecialDecision")
        item.reindexObject()
        self.assertEqual(len(self.catalog(SearchableText="SpecialDécision")), 1)
        self.assertEqual(len(self.catalog(SearchableText="specialdécision")), 1)
        self.assertEqual(len(self.catalog(SearchableText="SpecialDecision")), 1)
        self.assertEqual(len(self.catalog(SearchableText="specialdecision")), 1)
        # not found
        self.assertEqual(len(self.catalog(SearchableText="specialdecisions")), 0)

    def test_pm_Faceted_annexes_vocabulary(self):
        """Test especially that enabling attributes in various annexes_types works."""
        cfg = self.meetingConfig
        vocab = get_vocab(
            cfg,
            "Products.PloneMeeting.vocabularies.faceted_annexes_vocabulary",
            only_factory=True)
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        # nothing enabled, vocabulary is empty
        self.assertEqual(vocab(cfg)._terms, [])
        # enable "signed" for item_decision
        self._enable_annex_config(item, param="signed", related_to="item_decision")
        self.assertEqual(
            [term.token for term in vocab(cfg)],
            ['to_sign', 'not_to_sign', 'signed'])
        # enable "confidentiality" for item_annex
        self._enable_annex_config(item)
        self.assertEqual(
            [term.token for term in vocab(cfg)],
            ['confidential', 'not_confidential', 'to_sign', 'not_to_sign', 'signed'])

    def test_pm_CopyGroupsVocabulary(self):
        """Test, especially when using copyGroups and restrictedCopyGroups."""
        cfg = self.meetingConfig
        cfg.setSelectableCopyGroups((self.vendors_reviewers, self.developers_reviewers))
        cfg.setSelectableRestrictedCopyGroups((self.vendors_reviewers, ))
        self._enableField('copyGroups')
        self._enableField('restrictedCopyGroups')
        self.changeUser('pmManager')
        self.assertEqual(
            get_vocab_values(cfg, "Products.PloneMeeting.vocabularies.copygroupsvocabulary"),
            [self.developers_reviewers, self.vendors_reviewers])


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testFaceted, prefix='test_pm_'))
    return suite



# -*- coding: utf-8 -*-
#
# File: testPortlets.py
#
# Copyright (c) 2015 by Imio.be
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from plone.testing import layered
from Products.PloneMeeting.testing import PM_TESTING_ROBOT

import robotsuite


def test_suite():
    return layered(robotsuite.RobotTestSuite('robot/views.robot'),
                   layer=PM_TESTING_ROBOT)



# -*- coding: utf-8 -*-
import os


# attribute name on the portal to store keycloak auth token
AUTH_INFOS_ATTR = "keycloak_auth_infos"

# keycloak auth token configuration
VISION_AUTH_URL = os.getenv('VISION_AUTH_URL')
VISION_CLIENT_ID = os.getenv('VISION_CLIENT_ID')
VISION_CLIENT_SECRET = os.getenv('VISION_CLIENT_SECRET')
VISION_AUTH_USERNAME = os.getenv('VISION_AUTH_USERNAME')
VISION_AUTH_PASSWORD = os.getenv('VISION_AUTH_PASSWORD')
AUTH_CURL_COMMAND = "curl --location '%s' \
--header 'Content-Type: application/x-www-form-urlencoded' \
--header 'Cookie: KEYCLOAK_LOCALE=fr' \
--data-urlencode 'client_id=%s' \
--data-urlencode 'client_secret=%s' \
--data-urlencode 'username=%s' \
--data-urlencode 'password=%s' \
--data-urlencode 'grant_type=password'" % (
    VISION_AUTH_URL,
    VISION_CLIENT_ID,
    VISION_CLIENT_SECRET,
    VISION_AUTH_USERNAME,
    VISION_AUTH_PASSWORD)
VISION_MUNICIPALITY_ID = os.getenv('VISION_MUNICIPALITY_ID', '')

# api configuration
API_URL = 'https://ipa.imio.be/imio/vision/v1/municipalities/{0}/%s?delib_user=%s%s'.format(
    VISION_MUNICIPALITY_ID)



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from collections import OrderedDict
from collective.z3cform.datagridfield import DataGridFieldFactory
from collective.z3cform.datagridfield import DictRow
from imio.helpers.content import get_vocab
from imio.helpers.security import fplog
from persistent.mapping import PersistentMapping
from plone import api
from plone.autoform.directives import widget
from plone.restapi.deserializer import boolean_value
from Products.Five import BrowserView
from Products.PloneMeeting.browser.itemattendee import BaseAttendeeForm
from Products.PloneMeeting.browser.itemattendee import IBaseAttendee
from Products.PloneMeeting.config import NOT_ENCODED_VOTE_VALUE
from Products.PloneMeeting.config import NOT_VOTABLE_LINKED_TO_VALUE
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.utils import get_context_with_request
from Products.PloneMeeting.utils import notifyModifiedAndReindex
from Products.PloneMeeting.widgets.pm_number import PMNumberFieldWidget
from Products.PloneMeeting.widgets.pm_selectreadonly import PMSelectReadonlyWidget
from z3c.form import field
from z3c.form.browser.radio import RadioFieldWidget
from z3c.form.contentprovider import ContentProviders
from z3c.form.interfaces import HIDDEN_MODE
from z3c.form.interfaces import IFieldsAndContentProvidersForm
from zope import schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component.hooks import getSite
from zope.contentprovider.provider import ContentProviderBase
from zope.i18n import translate
from zope.interface import implements
from zope.interface import Interface
from zope.interface import Invalid
from zope.interface import invariant
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory


def _build_voting_groups(context, caching=True):
    """Build voting groups informations that will be used to manage the votes by group."""

    # caching is done in the REQUEST because this is called 2 times
    if caching and hasattr(context, "REQUEST"):
        res = getattr(context.REQUEST, '_build_voting_groups', None)

    if res is None:
        res = OrderedDict([('all', {'title': 'All', 'uids': []}),
                           ('others', {'title': 'Others', 'uids': []})])
        for voter in context.get_item_voters(theObjects=True):
            group_id = 'others'
            voting_group = voter.voting_group
            # use a voting_group if selected, else use 'others
            if voting_group and not voting_group.isBroken():
                org = voting_group.to_object
                group_id = org.getId()
                if group_id not in res:
                    res[group_id] = {'title': org.title, 'uids': []}
            res[group_id]['uids'].append(voter.UID())
        # only keep PLONEGROUP_ORG if any other value than 'all'
        if res.keys() == ['all', 'others']:
            res.pop('others')
        else:
            # reorder so PLONEGROUP_ORG is at the end
            ordered = res.keys()
            ordered += [ordered.pop(1)]
            res = OrderedDict((k, res[k]) for k in ordered)
        # caching
        if caching and hasattr(context, "REQUEST"):
            context.REQUEST.set('_build_voting_groups', res)
    return res


class DisplaySelectAllProvider(ContentProviderBase):
    """
      This ContentProvider will just display
      the select all controls to ease selecting same vote value for everyone.
    """
    template = \
        ViewPageTemplateFile('templates/display_select_all_vote_value.pt')

    def __init__(self, context, request, view):
        super(DisplaySelectAllProvider, self).__init__(
            context, request, view)
        self.__parent__ = view

    def render(self):
        used_vote_terms = get_vocab(
            self.context, "Products.PloneMeeting.vocabularies.usedvotevaluesvocabulary")
        self.used_vote_values = [term.token for term in used_vote_terms._terms]
        self.groups = _build_voting_groups(self.context)
        return self.template()


def vote_number_default():
    """
      Get the value from the REQUEST as it is passed when calling the
      form : form?vote_number=0.
    """
    request = getSite().REQUEST
    return request.get('vote_number', 0)


@provider(IContextAwareDefaultFactory)
def votes_default(context):
    """Default values for votes :
       - either we have itemVotes and we use it on voters;
       - or we do not have and we use the default value defined on MeetingConfig."""
    res = []
    vote_number = vote_number_default()
    item_votes = context.get_item_votes(
        vote_number=vote_number,
        ignored_vote_values=[NOT_VOTABLE_LINKED_TO_VALUE])
    # keep order using get_item_voters
    item_voter_uids = context.get_item_voters()
    # when adding a new vote linked_to_previous, only keep possible voters
    if item_votes['linked_to_previous']:
        is_new_vote = not context.get_item_votes(
            vote_number=vote_number, include_unexisting=False)
        if is_new_vote:
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(context)
            # only keep NOT_ENCODED_VOTE_VALUE
            ignored_vote_values = list(cfg.getUsedVoteValues())
            ignored_vote_values.append(NOT_VOTABLE_LINKED_TO_VALUE)
            last_vote = context.get_item_votes(
                ignored_vote_values=ignored_vote_values)[-1]
            item_voter_uids = [item_voter_uid for item_voter_uid in item_voter_uids
                               if item_voter_uid in last_vote['voters']]

    for item_voter_uid in item_voter_uids:
        # voter could not be there because ignored_vote_values
        if item_voter_uid in item_votes['voters']:
            data = {'voter_uid': item_voter_uid,
                    'voter': item_voter_uid,
                    'vote_value': item_votes['voters'][item_voter_uid]}
            res.append(data)
    return res


@provider(IContextAwareDefaultFactory)
def label_default(context):
    """ """
    res = None
    item_votes = context.get_item_votes(vote_number=vote_number_default())
    if item_votes and item_votes['label']:
        res = item_votes['label']
    return res


@provider(IContextAwareDefaultFactory)
def linked_to_previous_default(context):
    """ """
    request = getSite().REQUEST
    res = request.get('form.widgets.linked_to_previous', None)
    if res is None:
        item_votes = context.get_item_votes(vote_number=vote_number_default())
        res = item_votes['linked_to_previous']
    return res or False


class IVote(Interface):

    voter_uid = schema.Choice(
        title=u'Voter UID',
        required=False,
        vocabulary="Products.PloneMeeting.vocabularies.itemvotersvocabulary", )

    widget('voter', PMSelectReadonlyWidget)
    voter = schema.Choice(
        title=_(u"Voter"),
        required=False,
        vocabulary="Products.PloneMeeting.vocabularies.itemvotersvocabulary", )

    widget('vote_value', RadioFieldWidget)
    vote_value = schema.Choice(
        title=u'',   # no title as we use column header for select all vote values
        required=False,
        vocabulary="Products.PloneMeeting.vocabularies.usedvotevaluesvocabulary", )


class IEncodeVotes(IBaseAttendee):

    vote_number = schema.Int(
        title=_(u"Vote number"),
        description=_(u""),
        defaultFactory=vote_number_default,
        required=False)

    linked_to_previous = schema.Bool(
        title=_(u"Linked to previous"),
        description=_(u"This will link this vote with the previous one, "
                      u"if so, voter may only vote for one of the linked votes."),
        defaultFactory=linked_to_previous_default,
        required=False)

    votes = schema.List(
        title=u'Votes',
        value_type=DictRow(title=u'Votes', schema=IVote),
        defaultFactory=votes_default,
        required=True)

    label = schema.TextLine(
        title=_(u"Label"),
        description=_(u"Free label that will identify the vote, "
                      u"useful when several votes are defined on an item. "
                      u"Leave empty if not used."),
        defaultFactory=label_default,
        required=False)


def _get_linked_item_vote_numbers(context, meeting, vote_number=0):
    """Return every votes linked to given p_vote_number."""
    res = []
    item_votes = meeting.item_votes.get(context.UID())
    if item_votes:
        len_item_votes = len(item_votes)
        # while adding new secret vote, vote_number does still not exist
        if len(item_votes) > vote_number:
            origin_is_linked = item_votes[vote_number]['linked_to_previous']
        else:
            origin_is_linked = bool(context.REQUEST.form.get('form.widgets.linked_to_previous'))
        if origin_is_linked:
            # find first
            i = vote_number - 1
            while i and item_votes[i]['linked_to_previous']:
                res.append(i)
                i -= 1
            # i is now on the vote previous first 'linked_to_previous', keep it
            res.append(i)
        # not origin_is_linked only way is next or origin_is_linked, need to get next too
        res.append(vote_number)
        i = vote_number + 1
        while i < len_item_votes and item_votes[i]['linked_to_previous']:
            res.append(i)
            i += 1
        # not linked, return an empty list
        if res == [vote_number]:
            res = []
    return res


def clean_voters_linked_to(context, meeting, vote_number, new_voters):
    """ """
    linked_vote_numbers = _get_linked_item_vote_numbers(context, meeting, vote_number)
    if linked_vote_numbers:
        item_votes = meeting.item_votes[context.UID()]
        # clean other vote numbers
        # get values edited just now that will no more be useable on linked votes
        kept_voters = [voter_uid for voter_uid, voter_value in new_voters.items()
                       if voter_value not in [NOT_ENCODED_VOTE_VALUE, NOT_VOTABLE_LINKED_TO_VALUE]]
        for linked_vote_number in linked_vote_numbers:
            if linked_vote_number == vote_number:
                continue
            linked_item_vote = item_votes[linked_vote_number]
            for voter_uid, vote_value in linked_item_vote['voters'].items():
                if voter_uid in kept_voters:
                    linked_item_vote['voters'][voter_uid] = NOT_VOTABLE_LINKED_TO_VALUE
                elif vote_value == NOT_VOTABLE_LINKED_TO_VALUE and \
                        voter_uid in new_voters and \
                        new_voters[voter_uid] == NOT_ENCODED_VOTE_VALUE:
                    # this is potentially a liberated vote value
                    linked_item_vote['voters'][voter_uid] = NOT_ENCODED_VOTE_VALUE


def next_vote_is_linked(itemVotes, vote_number=0):
    """Return True if next vote is linked, useful when on first vote
       that has not the linked_to_previous info."""
    res = False
    if len(itemVotes) > vote_number + 1 and \
       itemVotes[vote_number + 1]['linked_to_previous']:
        res = True
    return res


def is_vote_updatable_for(context, item_to_update):
    """Given p_item_to_update will be updatable if it is the context or
       if using same pollType and same voters and not using several or linked votes."""
    res = False
    if context == item_to_update or \
       (item_to_update.getPollType() != 'no_vote' and
        context.get_item_votes(vote_number=0).get('poll_type', context.getPollType()) ==
        item_to_update.get_item_votes(vote_number=0).get('poll_type', item_to_update.getPollType()) and
        context.get_item_voters() == item_to_update.get_item_voters() and
            len(item_to_update.get_item_votes()) < 2):
        res = True
    return res


def display_item_numbers(numbers):
    """Manage displaying given p_elements with following result:
       - 1 numbers: "2";
       - 2 numbers: "2 & 3";
       - 5 numbers: "2, 3, 4, 5 & 6".
    """
    if len(numbers) > 1:
        res = ", ".join(numbers[:-1]) + " & " + numbers[-1]
    else:
        res = numbers[0]
    return res


class EncodeVotesForm(BaseAttendeeForm):
    """ """
    implements(IFieldsAndContentProvidersForm)
    contentProviders = ContentProviders()
    contentProviders['select_all'] = DisplaySelectAllProvider
    contentProviders['select_all'].position = 2

    label = _(u"Encode votes")
    schema = IEncodeVotes
    fields = field.Fields(IEncodeVotes)
    fields['votes'].widgetFactory = DataGridFieldFactory

    id = "item-encode-votes-form"

    def updateWidgets(self):
        # hide vote_number field
        self.fields['vote_number'].mode = HIDDEN_MODE
        # add a css class corresponding to group of held positions
        groups = _build_voting_groups(self.context)
        # do not hide it, when hidding it, value is always True???
        # this is hidden using CSS
        # self.fields['linked_to_previous'].mode = HIDDEN_MODE
        super(EncodeVotesForm, self).updateWidgets()
        self.widgets['votes'].allow_delete = False
        self.widgets['votes'].allow_insert = False
        self.widgets['votes'].allow_reorder = False
        self.widgets['votes'].auto_append = False
        self.widgets['votes'].columns[0]['mode'] = HIDDEN_MODE
        for row in self.widgets['votes'].widgets:
            if row.subform.context is None:
                continue
            voter_uid = row.subform.context['voter_uid']
            group_id = [group_id for group_id, values in groups.items()
                        if voter_uid in values['uids']]
            if group_id:
                row.addClass(group_id[0])
            for wdt in row.subform.widgets.values():
                if wdt.__name__ == 'voter_uid':
                    wdt.mode = HIDDEN_MODE
        # disable apply_until_item_number when using several votes
        # or if poll_type was redefined
        if _should_disable_apply_until_item_number(self.context):
            apply_until_item_number = self.widgets['apply_until_item_number']
            apply_until_item_number.disabled = "disabled"
            apply_until_item_number.title = _(
                u"Not available when using several votes on same item.")

    def _doApply(self):
        """ """
        error = super(EncodeVotesForm, self)._doApply()
        if error:
            return error

        # prepare Meeting.item_votes compatible data
        # while datagrid used in an overlay, some <NO_VALUE>
        # wipeout self.votes from these values
        self.votes = [vote for vote in self.votes if isinstance(vote, dict)]
        data = {}
        data['label'] = self.label
        data['linked_to_previous'] = self.linked_to_previous
        data['voters'] = PersistentMapping()
        for vote in self.votes:
            data['voters'][vote['voter_uid']] = vote['vote_value']
        # add poll_type in case it was redefined
        vote_infos = self.context.get_item_votes(vote_number=self.vote_number)
        if 'poll_type' in vote_infos:
            data['poll_type'] = vote_infos['poll_type']

        updated = []
        not_updated = []
        for item_to_update in self.items_to_update:
            # set item public votes
            if is_vote_updatable_for(self.context, item_to_update):
                self.meeting.set_item_public_vote(item_to_update, data, self.vote_number)
                notifyModifiedAndReindex(item_to_update)
                updated.append(item_to_update.getItemNumber(for_display=True))
            else:
                not_updated.append(item_to_update.getItemNumber(for_display=True))

        # finish
        voter_uids = [vote['voter_uid'] for vote in self.votes]
        voter_uids = "__".join(voter_uids)
        vote_values = [vote['vote_value'] for vote in self.votes]
        vote_values = "__".join(vote_values)
        first_item_number = self.items_to_update[0].getItemNumber(for_display=True)
        last_item_number = self.items_to_update[-1].getItemNumber(for_display=True)
        extras = 'item={0} vote_number={1} voter_uids={2} vote_values={3} ' \
            'from_item_number={4} until_item_number={5}'.format(
                repr(self.context),
                self.vote_number,
                voter_uids,
                vote_values,
                first_item_number,
                last_item_number)
        fplog('encode_item_votes', extras=extras)
        if len(updated) == 1:
            api.portal.show_message(
                _("Votes have been encoded for current item."),
                request=self.request)
        else:
            api.portal.show_message(
                _("Votes have been encoded for items \"${item_numbers}\".",
                  mapping={'item_numbers': display_item_numbers(updated)}),
                request=self.request)

        # display items that could not be updated
        if not_updated:
            api.portal.show_message(
                _("error_updating_votes_for_items",
                  mapping={'item_numbers': display_item_numbers(not_updated)}),
                request=self.request, type="warning")
        self._finished = True


@provider(IContextAwareDefaultFactory)
def secret_votes_default(context):
    """Default values for secret votes."""
    res = []
    vote_number = vote_number_default()
    item_votes = context.get_item_votes(vote_number=vote_number,
                                        ignored_vote_values=[NOT_VOTABLE_LINKED_TO_VALUE])
    used_vote_terms = get_vocab(
        context,
        "Products.PloneMeeting.vocabularies.usedvotevaluesvocabulary",
        vote_number=vote_number)
    usedVoteValues = [term.token for term in used_vote_terms._terms
                      if term.token != NOT_ENCODED_VOTE_VALUE]
    for usedVoteValue in usedVoteValues:
        data = {'vote_value_id': usedVoteValue,
                'vote_value': usedVoteValue,
                'vote_count': item_votes['votes'].get(usedVoteValue) or 0}
        res.append(data)
    return res


class ISecretVote(Interface):

    vote_value_id = schema.Choice(
        title=u'Vote value id',
        required=False,
        vocabulary="Products.PloneMeeting.vocabularies.usedvotevaluesvocabulary", )

    widget('vote_value', PMSelectReadonlyWidget)
    vote_value = schema.Choice(
        title=_(u"Vote value"),
        required=False,
        vocabulary="Products.PloneMeeting.vocabularies.usedvotevaluesvocabulary", )

    widget('vote_count', PMNumberFieldWidget)
    vote_count = schema.Int(
        title=_(u"Vote count"),
        required=False)


class IEncodeSecretVotes(IBaseAttendee):

    vote_number = schema.Int(
        title=_(u"Vote number"),
        description=_(u""),
        defaultFactory=vote_number_default,
        required=False)

    linked_to_previous = schema.Bool(
        title=_(u"Linked to previous"),
        description=_(u"This will link this vote with the previous one, "
                      u"if so, voter may only vote for one of the linked votes."),
        defaultFactory=linked_to_previous_default,
        required=False)

    votes = schema.List(
        title=_(u"Votes"),
        value_type=DictRow(title=u'Votes', schema=ISecretVote),
        defaultFactory=secret_votes_default,
        required=True)

    label = schema.TextLine(
        title=_(u"Label"),
        description=_(u"Free label that will identify the vote, "
                      u"useful when several votes are defined on an item. "
                      u"Leave empty if not used."),
        defaultFactory=label_default,
        required=False)

    @invariant
    def validate_votes(data):
        ''' '''
        # prepare Meeting.item_votes compatible data
        # while datagrid used in an overlay, some <NO_VALUE>
        # wipeout self.votes from these values
        # data.__context__ does not contain the context...
        context = data.__context__ or get_context_with_request(None)
        meeting = context.getMeeting()
        votes = [vote for vote in data.votes if isinstance(vote, dict)]
        data.votes = votes
        # when used in an overlay, PMNumberWidget number browser validation
        # is not done, so we do it here... wrong values are received as None
        none_values = [v for v in votes if v["vote_count"] is None]
        if none_values:
            msg = translate(u'error_some_values_are_not_integers',
                            domain="PloneMeeting",
                            context=context.REQUEST)
            raise Invalid(msg)
        # check if max voters of every linked secret votes is not exceeded
        linked_vote_numbers = _get_linked_item_vote_numbers(context, meeting, data.vote_number)
        max_voters = len(context.get_item_voters())
        if linked_vote_numbers:
            # init at current value
            total = sum([vote['vote_count'] for vote in data.votes])
            for linked_vote_number in linked_vote_numbers:
                if linked_vote_number != data.vote_number:
                    linked_vote = context.get_item_votes(vote_number=linked_vote_number)
                    total += sum([vote_count for vote_value, vote_count
                                  in linked_vote['votes'].items()])
        else:
            total = sum([vote['vote_count'] for vote in votes])
        if total > max_voters:
            msg = translate(u'error_can_not_encode_more_than_max_voters',
                            mapping={'max_voters': str(max_voters)},
                            domain="PloneMeeting",
                            context=context.REQUEST)
            raise Invalid(msg)


def _should_disable_apply_until_item_number(context):
    """Is it possible to use the "apply_until_item_number" field?"""
    vote_number = context.REQUEST.get('vote_number', 0)
    if len(context.get_item_votes()) > 1 or \
       vote_number > 0 or \
       context.get_item_votes(vote_number).get('poll_type', None) != \
       context.getPollType():
        return True
    return False


class EncodeSecretVotesForm(BaseAttendeeForm):
    """ """

    label = _(u"Encode secret votes")
    schema = IEncodeSecretVotes
    fields = field.Fields(IEncodeSecretVotes)
    fields['votes'].widgetFactory = DataGridFieldFactory

    def updateWidgets(self):
        # hide vote_number field
        self.fields['vote_number'].mode = HIDDEN_MODE
        # do not hide it, when hidding it, value is always True???
        # this is hidden using CSS
        # self.fields['linked_to_previous'].mode = HIDDEN_MODE
        super(EncodeSecretVotesForm, self).updateWidgets()
        self.widgets['votes'].allow_delete = False
        self.widgets['votes'].allow_insert = False
        self.widgets['votes'].allow_reorder = False
        self.widgets['votes'].auto_append = False
        self.widgets['votes'].columns[0]['mode'] = HIDDEN_MODE
        for row in self.widgets['votes'].widgets:
            for wdt in row.subform.widgets.values():
                if wdt.__name__ == 'vote_value_id':
                    wdt.mode = HIDDEN_MODE
                elif wdt.__name__ == 'vote_value' and wdt.value:
                    wdt.klass += " {0}".format(wdt.context['vote_value'])
        # disable apply_until_item_number when using several votes
        if _should_disable_apply_until_item_number(self.context):
            apply_until_item_number = self.widgets['apply_until_item_number']
            apply_until_item_number.disabled = "disabled"
            apply_until_item_number.title = _(u"Not available when using several votes on same item.")

    def max(self, widget):
        """ """
        return len(self.context.get_item_voters())

    def min(self, widget):
        """ """
        return 0

    def _doApply(self):
        """ """
        error = super(EncodeSecretVotesForm, self)._doApply()
        if error:
            return error

        # prepare Meeting.itemVotes compatible data
        # while datagrid used in an overlay, some <NO_VALUE>
        # wipeout self.votes from these values
        self.votes = [vote for vote in self.votes if isinstance(vote, dict)]
        data = {}
        data['label'] = self.label
        data['linked_to_previous'] = self.linked_to_previous
        data['votes'] = {}
        for vote in self.votes:
            data['votes'][vote['vote_value_id']] = vote['vote_count']

        updated = []
        not_updated = []
        for item_to_update in self.items_to_update:
            # set item secret vote
            if is_vote_updatable_for(self.context, item_to_update):
                self.meeting.set_item_secret_vote(item_to_update, data, self.vote_number)
                notifyModifiedAndReindex(item_to_update)
                updated.append(item_to_update.getItemNumber(for_display=True))
            else:
                not_updated.append(item_to_update.getItemNumber(for_display=True))

        # finish
        vote_values = [vote['vote_value_id'] for vote in self.votes]
        vote_values = "__".join(vote_values)
        vote_count = [str(vote['vote_count']) for vote in self.votes]
        vote_count = "__".join(vote_count)
        first_item_number = self.items_to_update[0].getItemNumber(for_display=True)
        last_item_number = self.items_to_update[-1].getItemNumber(for_display=True)
        extras = 'item={0} vote_number={1} vote_values={2} vote_count={3} ' \
            'from_item_number={4} until_item_number={5}'.format(
                repr(self.context),
                self.vote_number,
                vote_values,
                vote_count,
                first_item_number,
                last_item_number)
        fplog('encode_item_secret_votes', extras=extras)
        if len(updated) == 1:
            api.portal.show_message(
                _("Votes have been encoded for current item."),
                request=self.request)
        else:
            api.portal.show_message(
                _("Votes have been encoded for items \"${item_numbers}\".",
                  mapping={'item_numbers': display_item_numbers(updated)}),
                request=self.request)

        # display items that could not be updated
        if not_updated:
            api.portal.show_message(
                _("error_updating_votes_for_items",
                  mapping={'item_numbers': display_item_numbers(not_updated)}),
                request=self.request, type="warning")
        self._finished = True


class ItemDeleteVoteView(BrowserView):
    """ """

    def __call__(self, object_uid, redirect=True):
        """ """
        if not self.context._mayChangeAttendees():
            raise Unauthorized

        # redirect can by passed by jQuery, in this case, we receive '0' or '1'
        redirect = boolean_value(redirect)
        vote_number = int(object_uid)
        item_uid = self.context.UID()
        meeting = self.context.getMeeting()
        if item_uid in meeting.item_votes:
            item_votes = meeting.item_votes[item_uid]
            assert self.context._voteIsDeletable(meeting, vote_number)

            vote_to_delete = item_votes[vote_number]
            if self.context.get_vote_is_secret(meeting, vote_number):
                originnal_vote_keys = [str(vote_count) for vote_value, vote_count
                                       in vote_to_delete['votes'].items()]
                originnal_vote_keys = "__".join(originnal_vote_keys)
                originnal_vote_values = [vote_value for vote_value, vote_count
                                         in vote_to_delete['votes'].items()]
                originnal_vote_values = "__".join(originnal_vote_values)
                fp_extras_pattern = 'item={0} vote_number={1} vote_label={2} vote_count={3} vote_values={4}'
            else:
                originnal_vote_keys = [voter_uid for voter_uid in vote_to_delete['voters']]
                originnal_vote_keys = "__".join(originnal_vote_keys)
                originnal_vote_values = [vote_value for vote_value in vote_to_delete['voters'].values()]
                originnal_vote_values = "__".join(originnal_vote_values)
                fp_extras_pattern = 'item={0} vote_number={1} vote_label={2} voter_uids={3} vote_values={4}'
                # call clean_voters_linked_to with every values NOT_ENCODED_VOTE_VALUE
                # to liberate every values
                new_voters = vote_to_delete['voters'].copy()
                new_voters = {voter_uid: NOT_ENCODED_VOTE_VALUE
                              for voter_uid, vote_value in new_voters.items()
                              if vote_value != NOT_VOTABLE_LINKED_TO_VALUE}
                clean_voters_linked_to(self.context, meeting, vote_number, new_voters)

            # delete from meeting itemVote
            deleted_vote = meeting.item_votes[item_uid].pop(vote_number)
            # if deleted last existing vote (vote_number 0) remove context UID from meeting itemVotes
            if not meeting.item_votes[item_uid]:
                del meeting.item_votes[item_uid]
            # finish deletion
            label = deleted_vote['label'] or ""
            extras = fp_extras_pattern.format(
                repr(self.context),
                vote_number,
                label.encode('utf-8'),
                originnal_vote_keys,
                originnal_vote_values)
            fplog('delete_item_votes', extras=extras)
        # message
        api.portal.show_message(
            _("Votes number ${vote_number} have been deleted for current item.",
              mapping={'vote_number': vote_number + 1}),
            request=self.request)
        self._finished = True



# -*- coding: utf-8 -*-
#
# File: __init__.py
#
# GNU General Public License (GPL)
#

from AccessControl import allow_module
from AccessControl import allow_type
from datetime import datetime
from plone.registry.field import DisallowedProperty
from Products.Archetypes import listTypes
from Products.Archetypes.atapi import process_types
from Products.CMFCore import DirectoryView
from Products.CMFCore import utils as cmfutils
from Products.CMFPlone.utils import ToolInit
from Products.PloneMeeting.config import ADD_CONTENT_PERMISSIONS
from Products.PloneMeeting.config import DEFAULT_ADD_CONTENT_PERMISSION
from Products.PloneMeeting.config import product_globals
from Products.PloneMeeting.config import PROJECTNAME
from Products.validation import validation
from Products.validation.validators.BaseValidators import baseValidators
from Products.validation.validators.BaseValidators import protocols
from validators import ATCertifiedSignaturesValidator

import logging


DirectoryView.registerDirectory('skins', product_globals)

logger = logging.getLogger('PloneMeeting')
logger.debug('Installing Product')

__author__ = """Gaetan DELANNAY <gaetan.delannay@geezteem.com>, Gauthier BASTIEN
<g.bastien@imio.be>, Stephan GEULETTE <s.geulette@imio.be>"""
__docformat__ = 'plaintext'


# Another monkey patch in the "isURL" validator: why is the "file" protocol
# excluded?
protocols += ('file',)
for valid in baseValidators:
    if valid.name == 'isURL':
        del valid.regex[0]
        valid.regex_strings = (r'(%s)s?://[^\s\r\n]+' % '|'.join(protocols),)
        valid.compileRegex()

validation.register(ATCertifiedSignaturesValidator('isValidCertifiedSignatures', title='', description=''))


# this is necessary to be able to register custom validator for datagridfield
# we use it for validators.PloneGroupSettings validators,
# see https://github.com/collective/collective.z3cform.datagridfield/issues/14
DisallowedProperty('__provides__')


def initialize(context):
    """initialize product (called by zope)"""

    from Products.PloneMeeting import monkey
    import Meeting
    import MeetingCategory
    import MeetingConfig
    import MeetingGroup
    import MeetingItem
    import MeetingUser
    import ToolPloneMeeting

    # Initialize portal tools
    tools = [ToolPloneMeeting.ToolPloneMeeting]
    ToolInit(PROJECTNAME + ' Tools',
             tools=tools,
             icon='tool.gif').initialize(context)

    # Initialize portal content
    all_content_types, all_constructors, all_ftis = process_types(
        listTypes(PROJECTNAME),
        PROJECTNAME)

    cmfutils.ContentInit(
        PROJECTNAME + ' Content',
        content_types=all_content_types,
        permission=DEFAULT_ADD_CONTENT_PERMISSION,
        extra_constructors=all_constructors,
        fti=all_ftis).initialize(context)

    # Give it some extra permissions to control them on a per class limit
    for i in range(0, len(all_content_types)):
        klassname = all_content_types[i].__name__
        if klassname not in ADD_CONTENT_PERMISSIONS:
            continue

        context.registerClass(meta_type=all_ftis[i]['meta_type'],
                              constructors=(all_constructors[i],),
                              permission=ADD_CONTENT_PERMISSIONS[klassname])

    allow_module('collective.iconifiedcategory.safe_utils')
    allow_module('collective.contact.core.safe_utils')
    allow_module('collective.contact.plonegroup.safe_utils')
    allow_module('imio.annex.safe_utils')
    allow_module('imio.history.safe_utils')
    allow_module('Products.PloneMeeting.safe_utils')
    allow_module('Products.PloneMeeting.browser.meeting')
    allow_type(datetime)



# -*- coding: utf-8 -*-

from imio.helpers.xhtml import removeCssClasses
from plone import api
from plone.outputfilters.interfaces import IFilter
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.PloneMeeting.utils import isPowerObserverForCfg
from zope.interface import implements


class HighlightRemover(object):
    """Transform that will remove 'highlight' related CSS classes,
       depending on MeetingConfig.hideCSSClassesTo."""

    implements(IFilter)
    order = 1000

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    def is_enabled(self):
        """Hide if enabled when current user may not edit."""
        return bool(self.cfg) and \
            bool(self.cfg.getCssClassesToHide() and self.cfg.getHideCssClassesTo()) and \
            not _checkPermission(ModifyPortalContent, self.context)

    def _hideCssClasses(self, hideCssClassesTo):
        """Hide regarding MeetingConfig.hideCssClassesTo."""
        pos = [po["row_id"] for po in self.cfg.getPowerObservers()
               if po["row_id"] in hideCssClassesTo]
        if pos and \
           isPowerObserverForCfg(self.cfg, power_observer_types=pos):
            return True
        return False

    def __call__(self, data):
        hideCssClassesTo = self.cfg.getHideCssClassesTo()
        if hideCssClassesTo and self._hideCssClasses(hideCssClassesTo):
            return removeCssClasses(data, css_classes=self.cfg.getCssClassesToHide())
        return data



# -*- coding: utf-8 -*-

from z3c.form.browser.select import SelectWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import ISelectWidget
from z3c.form.widget import FieldWidget
from zope.interface import implementer
from zope.interface import implementer_only


class IPMSelectReadonlyWidget(ISelectWidget):
    """Marker interface for PM select readonly widget"""


@implementer_only(IPMSelectReadonlyWidget)
class PMSelectReadonlyWidget(SelectWidget):
    """Select readonly widget implementation."""


@implementer(IFieldWidget)
def PMSelectReadonlyFieldWidget(field, request):
    return FieldWidget(field, PMSelectReadonlyWidget(request))



# -*- coding: utf-8 -*-
#
# GNU General Public License (GPL)
#

from cgi import escape
from collections import OrderedDict
from eea.facetednavigation.interfaces import IFacetedNavigable
from plone import api
from plone.app.layout.viewlets import ViewletBase
from plone.app.layout.viewlets.common import TitleViewlet
from plone.memoize.view import memoize
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.PloneMeeting.config import ITEM_INITIATOR_INDEX_PATTERN
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.events import _is_held_pos_uid_used_by
from Products.PloneMeeting.utils import displaying_available_items
from zope.component import getMultiAdapter


class ForceInsertNormal(ViewletBase):
    '''This viewlet displays the forceInsertNormal button under the available
       items to present in a meeting on the meeting view.'''

    def available(self):
        """Always available on available items because we have there JS
           computing number of available items, we rely on the show method."""
        return displaying_available_items(self.context)

    def show(self):
        """ """
        return self.view.brains

    def enabled(self):
        """Is the checkbox enabled?  Only necessary if meeting is in a late state."""
        return self.context.is_late()

    def render(self):
        if self.available():
            return self.index()
        else:
            return ''

    def update(self):
        self.context_state = getMultiAdapter((self.context, self.request),
                                             name=u'plone_context_state')

    index = ViewPageTemplateFile("templates/viewlet_force_insert_normal.pt")


class HeldPositionBackRefs(ViewletBase):
    """Display elements using held_position."""

    index = ViewPageTemplateFile("templates/viewlet_held_position_back_refs.pt")

    def __init__(self, context, request, view, manager=None):
        super(HeldPositionBackRefs, self).__init__(context, request, manager)
        self.context_url = self.context.absolute_url()

    def available(self):
        """ """
        return True


class HeldPositionBackRefsView(BrowserView):
    """The asynch view that displays elements using held_position."""

    def using_configs(self):
        """ """
        tool = api.portal.get_tool('portal_plonemeeting')
        hp_uid = self.context.UID()
        res = []
        for cfg in tool.objectValues('MeetingConfig'):
            if _is_held_pos_uid_used_by(hp_uid, cfg):
                res.append(cfg)
        return res

    def using_meetings(self, limit=50):
        """ """
        tool = api.portal.get_tool('portal_plonemeeting')
        catalog = api.portal.get_tool('portal_catalog')
        hp_uid = self.context.UID()
        res = OrderedDict()
        for cfg in tool.objectValues('MeetingConfig'):
            meeting_type_name = cfg.getMeetingTypeName()
            brains = catalog.unrestrictedSearchResults(
                portal_type=meeting_type_name,
                sort_on='meeting_date',
                sort_order='reverse')
            for brain in brains:
                meeting = brain.getObject()
                if _is_held_pos_uid_used_by(hp_uid, meeting):
                    if cfg not in res:
                        res[cfg] = {'meetings': [], 'overlimit': False}
                    res[cfg]['meetings'].append(meeting)
                    if not limit or len(res[cfg]['meetings']) >= limit:
                        res[cfg]['overlimit'] = True
                        break
        return res

    def using_items(self, limit=50):
        """ """
        tool = api.portal.get_tool('portal_plonemeeting')
        catalog = api.portal.get_tool('portal_catalog')
        hp_uid = self.context.UID()
        res = OrderedDict()
        for cfg in tool.objectValues('MeetingConfig'):
            item_type_name = cfg.getItemTypeName()
            brains = catalog.unrestrictedSearchResults(
                portal_type=item_type_name,
                pm_technical_index=[
                    ITEM_INITIATOR_INDEX_PATTERN.format(hp_uid)],
                sort_on='meeting_date',
                sort_order='reverse')
            for brain in brains:
                if cfg not in res:
                    res[cfg] = {'items': [], 'overlimit': False}
                res[cfg]['items'].append(brain)
                if not limit or len(res[cfg]['items']) >= limit:
                    res[cfg]['overlimit'] = True
                    break
        return res


class PMTitleViewlet(TitleViewlet):

    @property
    @memoize
    def page_title(self):
        '''Include MeetingConfig title when on a faceted context (dashboards).'''
        title = super(PMTitleViewlet, self).page_title
        if IFacetedNavigable.providedBy(self.context) and \
           not IMeeting.providedBy(self.context):
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self.context)
            if cfg:
                title = u"%s - %s" % (escape(safe_unicode(cfg.Title())), title)
        return title



# -*- coding: utf-8 -*-

from collective.contact.plonegroup.utils import get_own_organization
from collective.eeafaceted.dashboard.browser.facetedcollectionportlet import Renderer as FacetedRenderer
from eea.facetednavigation.interfaces import IFacetedNavigable
from plone.app.portlets.portlets import base
from plone.portlets.interfaces import IPortletDataProvider
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.PloneMeeting.content.meeting import IMeeting
from zope.formlib import form
from zope.i18nmessageid import MessageFactory
from zope.interface import implements


_ = MessageFactory('PloneMeeting')


class IPloneMeetingPortlet(IPortletDataProvider):
    """
      A portlet that shows controls of PloneMeeting
    """


class Assignment(base.Assignment):
    implements(IPloneMeetingPortlet)

    def __init__(self):
        pass

    @property
    def title(self):
        return _(u"PloneMeeting")


class Renderer(base.Renderer, FacetedRenderer):

    _template = ViewPageTemplateFile('templates/portlet_plonemeeting.pt')

    def __init__(self, *args):
        base.Renderer.__init__(self, *args)
        self.portal = getToolByName(self.context, 'portal_url').getPortalObject()
        self.portal_url = self.portal.absolute_url()
        self.tool = getToolByName(self.portal, 'portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    @property
    def available(self):
        '''Defines if the portlet is available in the context.'''
        available = FacetedRenderer(self.context, self.request, self.view, self.manager, self.data).available
        return available

    @property
    def _criteriaHolder(self):
        '''Override to not consider IMeeting as a criteria holder.'''
        parent = self.context
        # look up parents until we found the criteria holder or we reach the 'Plone Site'
        while parent and not parent.portal_type == 'Plone Site':
            if IFacetedNavigable.providedBy(parent) and not IMeeting.providedBy(parent):
                return parent
            parent = parent.aq_inner.aq_parent

    def render(self):
        return self._template()

    def getPloneMeetingFolder(self):
        '''Returns the current PM folder.'''
        if self.cfg:
            return self.tool.getPloneMeetingFolder(self.cfg.getId())

    def get_own_org(self):
        ''' '''
        return get_own_organization()


class AddForm(base.AddForm):
    form_fields = form.Fields(IPloneMeetingPortlet)
    label = _(u"Add PloneMeeting Portlet")
    description = _(u"This portlet shows controls of PloneMeeting.")

    def create(self, data):
        return Assignment(**data)


class EditForm(base.EditForm):
    form_fields = form.Fields(IPloneMeetingPortlet)
    label = _(u"Edit PloneMeeting Portlet")
    description = _(u"This portlet shows controls of PloneMeeting.")



# -*- coding: utf-8 -*-
#
# File: columns.py
#
# GNU General Public License (GPL)
#

from collective.contact.plonegroup.browser.tables import OrgaPrettyLinkWithAdditionalInfosColumn
from collective.eeafaceted.z3ctable.columns import AbbrColumn
from collective.eeafaceted.z3ctable.columns import ActionsColumn
from collective.eeafaceted.z3ctable.columns import BaseColumn
from collective.eeafaceted.z3ctable.columns import BrowserViewCallColumn
from collective.eeafaceted.z3ctable.columns import CheckBoxColumn
from collective.eeafaceted.z3ctable.columns import ColorColumn
from collective.eeafaceted.z3ctable.columns import DateColumn
from collective.eeafaceted.z3ctable.columns import I18nColumn
from collective.eeafaceted.z3ctable.columns import PrettyLinkColumn
from collective.eeafaceted.z3ctable.columns import VocabularyColumn
from imio.annex.columns import ActionsColumn as AnnexActionsColumn
from imio.dashboard.interfaces import IContactsDashboard
from imio.prettylink.interfaces import IPrettyLink
from plone import api
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.PloneMeeting.config import AddAnnex
from Products.PloneMeeting.config import AddAnnexDecision
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.utils import displaying_available_items
from zope.i18n import translate


class ItemCategoryColumn(VocabularyColumn):
    """A column that display the MeetingItem.category."""
    vocabulary = u'Products.PloneMeeting.vocabularies.categoriesvocabulary'

    def getCSSClasses(self, item):
        """Add a class with category id so we can skin particular categories."""
        css_classes = self.cssClasses
        css_classes['td'] = css_classes['td'] + ' td_cell_{0}'.format(item.getCategory)
        return css_classes


class MeetingCategoryColumn(ItemCategoryColumn):
    """A column that display the Meeting.category."""
    attrName = 'getCategory'
    vocabulary = u'Products.PloneMeeting.vocabularies.meeting_categories_vocabulary'


class ItemClassifierColumn(VocabularyColumn):
    """A column that display the MeetingItem.classifier."""
    vocabulary = u'Products.PloneMeeting.vocabularies.classifiersvocabulary'


class ItemProposingGroupColumn(VocabularyColumn):
    """A column that display the MeetingItem.proposingGroup."""
    vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsvocabulary'
    header_help = u'header_getProposingGroup_help'


class ItemProposingGroupAcronymColumn(AbbrColumn):
    """A column that display the MeetingItem.proposingGroup acronym."""
    attrName = 'getProposingGroup'
    vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsacronymsvocabulary'
    full_vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsvocabulary'
    header_help = u'header_proposing_group_acronym_help'


class ItemGroupsInChargeColumn(VocabularyColumn):
    """A column that display the groupsInCharge."""
    attrName = 'getGroupsInCharge'
    vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsvocabulary'


class ItemGroupsInChargeAcronymColumn(AbbrColumn):
    """A column that display the groupsInCharge acronym."""
    attrName = 'getGroupsInCharge'
    vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsacronymsvocabulary'
    full_vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsvocabulary'
    header_help = u"header_groups_in_charge_acronym_help"


class ItemCopyGroupsColumn(VocabularyColumn):
    """A column that display the copyGroups."""
    attrName = 'getAllBothCopyGroups'
    vocabulary = u'Products.PloneMeeting.Groups'
    the_object = True


class ItemAssociatedGroupsColumn(VocabularyColumn):
    """A column that display the associatedGroups."""
    attrName = 'getAssociatedGroups'
    vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsvocabulary'
    header_help = u"header_getAssociatedGroups_help"


class ItemAssociatedGroupsAcronymColumn(AbbrColumn):
    """A column that display the associatedGroups acronym."""
    attrName = 'getAssociatedGroups'
    vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsacronymsvocabulary'
    full_vocabulary = u'Products.PloneMeeting.vocabularies.everyorganizationsvocabulary'
    header_help = u"header_associated_groups_acronym_help"


class ItemCommitteesColumn(VocabularyColumn):
    """A column that display the MeetingItem.committees."""
    attrName = 'committees_index'
    vocabulary = u'Products.PloneMeeting.vocabularies.selectable_committees_vocabulary'


class ItemCommitteesAcronymColumn(AbbrColumn):
    """A column that display the MeetingItem.committees acronym."""
    attrName = 'committees_index'
    vocabulary = u'Products.PloneMeeting.vocabularies.selectable_committees_acronyms_vocabulary'
    full_vocabulary = u'Products.PloneMeeting.vocabularies.selectable_committees_vocabulary'
    header_help = u"header_committees_index_acronym_help"


class ItemAdvicesColumn(BrowserViewCallColumn):
    """A column that display the MeetingItem advices."""
    view_name = 'advices-icons'
    escape = False


class ItemToDiscussColumn(BrowserViewCallColumn):
    """A column that display the MeetingItem.toDiscuss as an icon."""
    view_name = 'item-to-discuss'
    header_image = 'toDiscussYes.png'
    escape = False


class ItemIsSignedColumn(BrowserViewCallColumn):
    """A column that display the MeetingItem.toDiscuss as an icon."""
    view_name = 'item-is-signed'
    header_image = 'itemIsSignedYes.png'
    escape = False


class ItemPrivacyColumn(I18nColumn):
    """A column that display the translated MeetingItem.privacy."""
    i18n_domain = 'PloneMeeting'
    header_help = u'header_privacy_help'
    escape = False


class ItemPollTypeColumn(VocabularyColumn):
    """A column that display the MeetingItem.pollType."""
    vocabulary = u'Products.PloneMeeting.vocabularies.polltypesvocabulary'
    header_help = u'header_pollType_help'
    escape = False


def render_item_annexes(item, tool, show_nothing=False, check_can_view=False):
    """ """
    annexes = ''
    annexes += item.restrictedTraverse('@@categorized-childs')(
        portal_type='annex', show_nothing=show_nothing, check_can_view=check_can_view)
    decision_annexes = item.restrictedTraverse('@@categorized-childs')(
        portal_type='annexDecision', show_nothing=show_nothing, check_can_view=check_can_view)
    if decision_annexes.strip():
        decision_term = translate("AnnexesDecisionShort",
                                  domain='PloneMeeting',
                                  context=item.REQUEST)
        annexes += u"<span class='discreet'>{0}&nbsp;:&nbsp;</span>".format(decision_term)
        annexes += decision_annexes
    return annexes


class PMPrettyLinkColumn(PrettyLinkColumn):
    """A column that display the IPrettyLink.getLink column."""

    def renderHeadCell(self):
        """Override rendering of head of the cell to include jQuery
           call to initialize annexes menu and to show the 'more/less details' if we are listing items."""

        if not self.header_js:
            # avoid problems while concataining None and unicode
            self.header_js = u''
        # activate necessary javascripts
        self.header_js += u'<script type="text/javascript">initializeDashboard();' + \
            u'initializeIconifiedCategoryWidget();</script>'

        if self.table.batch and self.table.batch[0].meta_type == 'MeetingItem':
            # change header title to "Purpose"
            self.header = "header_purpose"
            showHideMsg = translate("show_or_hide_details",
                                    domain="PloneMeeting",
                                    context=self.request,
                                    default="Show/hide details")
            header = \
                u'<span class="showHideDetails" onclick="javascript:toggleMeetingDescriptions()">' + \
                u'<img src="{0}/more_less_details.png" title="{1}" /></span>'.format(
                    self.table.portal_url, showHideMsg)
            return super(PMPrettyLinkColumn, self).renderHeadCell() + header
        return super(PMPrettyLinkColumn, self).renderHeadCell()

    def renderCell(self, item):
        """ """
        obj = self._getObject(item)
        prettyLinker = IPrettyLink(obj)
        prettyLinker.target = '_parent'
        prettyLinker.showContentIcon = True

        annexes = staticInfos = moreInfos = ''

        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        if obj.meta_type == 'MeetingItem':
            isPrivacyViewable = obj.adapted().isPrivacyViewable()
            prettyLinker.isViewable = isPrivacyViewable
            # display annexes and infos if item and item isPrivacyViewable
            if isPrivacyViewable:
                # display moreInfos about item
                # visible columns are one define for items listings or when the meeting is displayed
                # so check where we are
                if IMeeting.providedBy(self.context):
                    # on the meeting
                    if displaying_available_items(self.context):
                        visibleColumns = cfg.getAvailableItemsListVisibleColumns()
                    else:
                        visibleColumns = cfg.getItemsListVisibleColumns()
                else:
                    visibleColumns = cfg.getItemColumns()
                staticInfos = obj.restrictedTraverse('@@static-infos')(
                    visibleColumns=visibleColumns)
                if self.showMoreInfos:
                    moreInfos = obj.restrictedTraverse('@@item-more-infos')()

                # display annexes
                annexes = render_item_annexes(obj, tool)
        elif obj.getTagName() == 'Meeting':
            staticInfos = obj.restrictedTraverse('@@static-infos')(
                visibleColumns=cfg.getMeetingColumns())
            # check_can_view=True because permission check is not enough
            annexes += obj.restrictedTraverse('@@categorized-childs')(
                portal_type='annex', check_can_view=True)
            # display number of items in meeting title
            contentValue = "{0} [{1}]".format(
                obj.Title(), obj.number_of_items())
            prettyLinker.contentValue = contentValue

        if annexes:
            annexes = u"<div class='dashboard_annexes'>{0}</div>".format(annexes)

        pretty_link = prettyLinker.getLink()
        return pretty_link + staticInfos + moreInfos + annexes

    @property
    def showMoreInfos(self):
        """ """
        return self.request.cookies.get('pmShowDescriptions', 'false') == 'true' and True or False

    def getCSSClasses(self, item):
        """Apply a particular class on the table row depending on the item's privacy
           if item is displayed in a meeting."""
        css_classes = super(PMPrettyLinkColumn, self).getCSSClasses(item)
        if self.context.getTagName() == 'Meeting':
            # for TR
            trCSSClasses = []
            trCSSClasses.append('meeting_item_privacy_{0}'.format(item.privacy))
            css_classes.update({'tr': ' '.join(trCSSClasses), })
        return css_classes


class PMActionsColumn(ActionsColumn):
    """A column displaying available actions of the listed item."""

    def renderCell(self, item):
        # dashboard displaying contacts
        if IContactsDashboard.providedBy(self.context):
            self.params['showActions'] = False
        return super(ActionsColumn, self).renderCell(item)


class PMAsyncActionsColumn(BaseColumn):
    """A column that displays an icon that will on hover display a tooltipster
       with actions availble to current user."""

    sort_index = -1
    escape = False

    def renderCell(self, item):
        tag = """<div class="async-actions-panel-icon-container">""" \
            """<a href="#" title="Actions" onclick="event.preventDefault();;">""" \
            """<span class="fa async-actions-panel-icon link-action tooltipster-actions-panel" data-base_url="{}" """ \
            """data-showHistory:boolean="1" /></a></div>""".format(item.getURL())
        return tag


class ItemLinkedMeetingColumn(BaseColumn):
    """
      Display the formatted date and a link to the linked meeting if any.
    """
    meeting_uid_attr = 'meeting_uid'
    attrName = 'meeting_date'
    use_caching = True
    escape = False

    def renderCell(self, item):
        """ """
        value = self.getValue(item)
        if self.use_caching:
            res = self._get_cached_result(value)
            if res:
                return res

        if not value or value.year <= 1950:
            res = u'-'
        else:
            catalog = self.table.portal.portal_catalog
            # done unrestricted because can be used to display meeting date
            # in dashboard when current user may not see the meeting
            brains = catalog.unrestrictedSearchResults(
                UID=getattr(item, self.meeting_uid_attr))
            meeting = brains[0]._unrestrictedGetObject()
            res = meeting.get_pretty_link()

        if self.use_caching:
            self._store_cached_result(value, res)
        return res


class ItemPreferredMeetingColumn(ItemLinkedMeetingColumn):
    """
      Display the formatted date and a link to the preferred meeting if any.
    """
    meeting_uid_attr = 'preferred_meeting_uid'
    attrName = 'preferred_meeting_date'
    header_help = u'header_preferred_meeting_date_help'
    escape = False


class ItemListTypeColumn(VocabularyColumn, ColorColumn):
    """A column that display the MeetingItem.listType as a color.
       We use a mix of VocabularyColumn and ColorColumn."""
    i18n_domain = "PloneMeeting"
    cssClassPrefix = 'meeting_item'
    vocabulary = u'Products.PloneMeeting.vocabularies.listtypesvocabulary'
    # VocabularyColumn and ColorColumn manage escape manually
    escape = False

    def renderCell(self, item):
        """Display a message."""
        term_value = super(ItemListTypeColumn, self).renderCell(item)
        # display the menu to change listType if current user may edit the meeting
        if _checkPermission(ModifyPortalContent, self.context):
            obj = self._getObject(item)
            return u'<div title="{0}" class="item_listType_container">' \
                u'<div class="item_listType Editable pmAction tooltipster-item-listtype-change" ' \
                u'data-base_url="{1}"></div></div>'.format(
                    term_value, obj.absolute_url())
        else:
            return u'<div title="{0}" class="item_listType_container"></div>'.format(term_value)


class ItemNumberColumn(BrowserViewCallColumn):
    """
      Display the itemNumber column, used on meetings.
    """
    view_name = 'item-number'
    header_js = u'<script type="text/javascript">initializeMeetingItemsDND();</script>'
    escape = False

    @property
    def cssClasses(self):
        """ """
        cssClasses = super(ItemNumberColumn, self).cssClasses
        if self.may_change_items_order(self.context):
            cssClasses['th'] = 'th_header_draggable'
            cssClasses['td'] = 'draggable'
        return cssClasses

    def may_change_items_order(self, meeting):
        """ """
        _may_change_items_order = getattr(self.table, '_may_change_items_order', None)
        if _may_change_items_order is None:
            self.table._may_change_items_order = meeting.wfConditions().may_change_items_order()
            if not self.table._may_change_items_order:
                # avoid init table DND, would slow browser for nothing
                self.header_js = u''
        return self.table._may_change_items_order

    def renderHeadCell(self):
        """ """
        cell = super(ItemNumberColumn, self).renderHeadCell()
        if self.may_change_items_order(self.context):
            cell = u'</th><th class="th_header_getItemNumber">' + cell
        return cell

    def renderCell(self, item):
        """ """
        may_change_items_order = self.may_change_items_order(self.context)
        self.params = {'may_change_items_order': may_change_items_order}
        cell = super(ItemNumberColumn, self).renderCell(item)
        if may_change_items_order:
            obj = self._getObject(item)
            cell = u"⣿</td><td td_cell_getItemNumber data-item_number='{0}'>".format(
                obj.getItemNumber()) + cell
        return cell


class ItemCheckBoxColumn(CheckBoxColumn):
    """ """

    def show_insert_or_remove_selected_items_action(self):
        '''Return True/False if the 'Remove selected items' or 'Present selected items'
           action must be displayed on the meeting view displaying presented items.'''
        return bool(_checkPermission(ModifyPortalContent, self.context) and
                    not self.context.query_state() in self.context.MEETINGCLOSEDSTATES)

    def renderHeadCell(self):
        """Display the '(un)present every selected items' action depending
           on the faceted we are on, available or presented items."""
        head = super(ItemCheckBoxColumn, self).renderHeadCell()
        if self.context.getTagName() == 'Meeting':
            if self.show_insert_or_remove_selected_items_action():
                if displaying_available_items(self.context):
                    present_msg = translate('present_several_items',
                                            domain='PloneMeeting',
                                            context=self.request)
                    head = u'''<table class="actionspanel-no-style-table nosort">
<tr><td>{0}</td><td><button onclick="presentSelectedItems('{1}')" title="{2}" class="present_several" type="button">
<img src="{3}/presentSeveral.png">
</button></td></tr></table>'''.format(head, self.context.absolute_url(), present_msg, self.table.portal_url)
                else:
                    unpresent_msg = translate('remove_several_items',
                                              domain='PloneMeeting',
                                              context=self.request)
                    head = u'''<table class="actionspanel-no-style-table nosort">
<tr><td>{0}</td><td><button onclick="removeSelectedItems('{1}')" title="{2}" class="remove_several" type="button">
<img src="{3}/removeSeveral.png">
</button></td></tr></table>'''.format(head, self.context.absolute_url(), unpresent_msg, self.table.portal_url)
        return head


class PMAnnexActionsColumn(AnnexActionsColumn):
    """ """
    params = AnnexActionsColumn.params.copy()
    params.update({'edit_action_class': 'link-overlay-pm-annex'})
    params.update({'arrowsPortalTypeAware': True})

    def renderCell(self, item):
        # display arrows for annex and/or annexDecision depending on relevant add permissions
        obj = self._getObject(item)
        parent = obj.aq_parent
        if not self._showArrows() or \
           (obj.portal_type == 'annex' and not _checkPermission(AddAnnex, parent)):
            self.params['showArrows'] = False
        elif not self._showArrows() or \
                (obj.portal_type == 'annexDecision' and not _checkPermission(AddAnnexDecision, parent)):
            self.params['showArrows'] = False
        else:
            self.params['showArrows'] = True
        return super(AnnexActionsColumn, self).renderCell(item)


class ReviewStateTitle(I18nColumn):
    """Translate the review_state title and not the id."""

    def _get_workflow(self, item):
        ''' '''
        if not hasattr(self, '_cached_wf'):
            wfTool = api.portal.get_tool('portal_workflow')
            wf = wfTool.getWorkflowsFor(item.portal_type)[0]
            self._cached_wf = wf
        return self._cached_wf

    def getValue(self, item):
        """ """
        wf = self._get_workflow(item)
        return wf.states.get(item.review_state).title


class PMOrgaPrettyLinkWithAdditionalInfosColumn(OrgaPrettyLinkWithAdditionalInfosColumn):
    """ """

    ai_reloaded_fields = ['position_type']


class ItemMeetingDeadlineDateColumn(DateColumn):
    """ """
    sort_index = -1
    attrName = "meetingDeadlineDate"
    long_format = True
    the_object = True



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from imio.helpers.cache import get_cachekey_volatile
from imio.helpers.cache import get_current_user_id
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.cache import invalidate_cachekey_volatile_for
from imio.helpers.content import get_user_fullname
from imio.helpers.content import get_vocab
from imio.helpers.content import uuidToObject
from plone import api
from plone.memoize import ram
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.PloneMeeting.browser.itemvotes import _get_linked_item_vote_numbers
from Products.PloneMeeting.browser.meeting import BaseMeetingView
from Products.PloneMeeting.config import NOT_ENCODED_VOTE_VALUE
from Products.PloneMeeting.config import NOT_VOTABLE_LINKED_TO_VALUE
from Products.PloneMeeting.config import WriteBudgetInfos
from Products.PloneMeeting.utils import reindex_object
from Products.PloneMeeting.utils import sendMailIfRelevant
from zope.i18n import translate


class Discuss(BrowserView):

    IMG_TEMPLATE = u'<img class="toDiscussEditable" src="%s" title="%s" name="%s" />'

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = api.portal.get()

    def isAsynchToggleEnabled(self):
        """
          Return True if the asynchronous call is enabled.
          Indeed, in some case, it can be necessary to deactivate the asynchronous call
          so the page is reloaded entirely if some other area are updated after
          having toggled the toDiscuss value.
        """
        return True

    def asynchToggle(self):
        """
          Toggle the MeetingItem.toDiscuss attribute from True to False and False to True
          asynchronously, meaning that the entire page is not fully reloaded but only the
          relevant icon.
        """
        if not _checkPermission(ModifyPortalContent, self.context):
            raise Unauthorized

        toDiscuss = not self.context.getToDiscuss()
        self.context.setToDiscuss(toDiscuss)
        self.context.adapted().onDiscussChanged(toDiscuss)

        if toDiscuss:
            filename = 'toDiscussYes.png'
            name = 'discussNo'
            title_msgid = 'to_discuss_yes_edit'
        else:
            filename = 'toDiscussNo.png'
            name = 'discussYes'
            title_msgid = 'to_discuss_no_edit'

        title = self.context.translate(title_msgid,
                                       domain="PloneMeeting")

        portal_url = self.portal.absolute_url()
        src = "%s/%s" % (portal_url, filename)

        html = self.IMG_TEMPLATE % (src, title, name)
        self.context._update_after_edit(idxs=['to_discuss'])
        return html

    def synchToggle(self, itemUid, discussAction):
        """
          This is a synchronous way of toggling toDiscuss.
          The asynchronous asynchToggle here above will only reload the clicked icon.
          If for some reason it is necessary that the page is fully reloaded,
          like for example to display a portal_message or because something else
          has changed on the page, this is the method to use.
          Here, it manages for example the fact that a reviewer can ask
          an item to be discussed and that will display a portal_message to this user.
        """
        item = uuidToObject(itemUid, unrestricted=True)
        if discussAction == 'ask':
            # I must send a mail to MeetingManagers for notifying them that a reviewer
            # wants to discuss this item.
            sendMailEnabled = sendMailIfRelevant(item, 'askDiscussItem', 'meetingmanagers', isSuffix=True)
            if sendMailEnabled:
                msgId = 'to_discuss_ask_mail_sent'
            else:
                msgId = 'to_discuss_ask_mail_not_sent'
            self.context.plone_utils.addPortalMessage(
                item.translate(msgId, domain='PloneMeeting'))
        elif discussAction == 'toggle':
            # I must toggle the "toDiscuss" switch on the item
            toDiscuss = not item.getToDiscuss()
            item.setToDiscuss(toDiscuss)
            item.adapted().onDiscussChanged(toDiscuss)
        self.context._update_after_edit(idxs=['to_discuss'])
        return self.request.RESPONSE.redirect(self.request['HTTP_REFERER'])


class TakenOverBy(BrowserView):
    """
      View that switch the item 'takenOverBy' from None to current user and from current user to None.
    """
    IMG_TEMPLATE = u'<span class="takenOverByEditable %s" title="%s" name="%s">\n%s\n</span>'

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = api.portal.get()

    def toggle(self, takenOverByFrom):
        if not self.context.adapted().mayTakeOver():
            raise Unauthorized

        memberId = get_current_user_id()

        currentlyTakenOverBy = self.context.getTakenOverBy()
        if currentlyTakenOverBy and \
           not currentlyTakenOverBy == takenOverByFrom and \
           not currentlyTakenOverBy == memberId:
            plone_utils = api.portal.get_tool('plone_utils')
            plone_utils.addPortalMessage(
                self.context.translate("The item you tried to take over was already taken "
                                       "over in between by ${fullname}. You can take it over "
                                       "now if you are sure that the other user do not handle it.",
                                       mapping={'fullname': get_user_fullname(currentlyTakenOverBy)},
                                       domain="PloneMeeting"),
                type='warning')
            self.request.RESPONSE.status = 500
            return

        # toggle value
        if not currentlyTakenOverBy:
            self.context.setTakenOverBy(memberId)
            newlyTakenOverBy = memberId
        else:
            self.context.setTakenOverBy('')
            newlyTakenOverBy = ''

        css_class = 'takenOverByNobody'
        name = 'takenOverByNo'
        title_msgid = 'taken_over_by_no_edit'
        if newlyTakenOverBy:
            if memberId == newlyTakenOverBy:
                css_class = 'takenOverByCurrentUser'
            else:
                css_class = 'takenOverByOtherUser'
            name = 'takenOverByYes'
            title_msgid = 'taken_over_by_yes_edit'

        title = translate(title_msgid,
                          domain="PloneMeeting",
                          context=self.request)

        if newlyTakenOverBy:
            taken_over_by = translate('Taken over by ${fullname}',
                                      mapping={'fullname': get_user_fullname(memberId)},
                                      domain="PloneMeeting",
                                      default="Taken over by ${fullname}",
                                      context=self.request)
        else:
            taken_over_by = translate('(Nobody)',
                                      domain="PloneMeeting",
                                      default="(Nobody)",
                                      context=self.request)

        html = self.IMG_TEMPLATE % (css_class, title, name, taken_over_by)
        # do not notifyModifiedAndReindex because an item may be taken over
        # when it is decided, by members of the proposingGroup
        # and in this case item must not be modified
        # cache will be invalidated because we check for modified and _p_mtime
        reindex_object(self.context, idxs=['getTakenOverBy'], update_metadata=False)
        return html


class AdviceIsConfidential(BrowserView):
    """
      View that switch an advice 'isConfidential' attribute using an ajax call.
    """
    IMG_TEMPLATE = u'<img src="%s" title="%s" name="%s" />'

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = api.portal.get()

    def toggle(self, UID):
        # get the adviceId
        adviceId = UID.split('__')[1]

        # check if current user may edit advice confidentiality
        if not self.context.adapted().mayEditAdviceConfidentiality(adviceId):
            raise Unauthorized

        isConfidential = self.context.adviceIndex[adviceId]['isConfidential']
        # toggle value
        self.context.adviceIndex[adviceId]['isConfidential'] = not isConfidential

        if isConfidential:
            filename = 'isConfidentialNo.png'
            name = 'isConfidentialYes'
            title_msgid = 'advice_is_confidential_no_edit'
        else:
            filename = 'isConfidentialYes.png'
            name = 'isConfidentialNo'
            title_msgid = 'advice_is_confidential_yes_edit'

        title = translate(title_msgid,
                          domain="PloneMeeting",
                          context=self.request)
        portal_url = self.portal.absolute_url()
        src = "%s/%s" % (portal_url, filename)
        html = self.IMG_TEMPLATE % (src, title, name)
        self.context._update_after_edit(idxs=[])
        return html


class BudgetRelated(BrowserView):
    """
      View that switch the item 'budgetRelated' attribute using an ajax call.
    """
    IMG_TEMPLATE = u'<img class="budgetRelatedEditable" src="%s" name="%s" title="%s" />' \
                   u' <span class="discreet %s">%s</span>'

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = api.portal.get()

    def toggle(self):
        if not _checkPermission(WriteBudgetInfos, self.context):
            raise Unauthorized

        beforeToggleBudgetRelated = self.context.getBudgetRelated()
        # toggle value
        self.context.setBudgetRelated(not beforeToggleBudgetRelated)

        if beforeToggleBudgetRelated:
            filename = 'budgetRelatedNo.png'
            # prefix with 'name' so we can discriminate this label from icon name
            name = 'nameBudgetRelatedYes'
            msgid = 'budget_related_no_edit'
            img_title_msgid = 'budget_related_no_img_title_edit'
        else:
            filename = 'budgetRelatedYes.png'
            name = 'nameBudgetRelatedNo'
            msgid = 'budget_related_yes_edit'
            img_title_msgid = 'budget_related_yes_img_title_edit'

        label = translate(msgid,
                          domain="PloneMeeting",
                          context=self.request)
        img_title = translate(img_title_msgid,
                              domain="PloneMeeting",
                              context=self.request)
        portal_url = self.portal.absolute_url()
        src = "%s/%s" % (portal_url, filename)
        budgetRelatedClass = beforeToggleBudgetRelated and 'notBudgetRelated' or 'budgetRelated'
        html = self.IMG_TEMPLATE % (src, name, img_title, budgetRelatedClass, label)
        # reload the page if current toggle did change adviceIndex
        # indeed, budgetRelated informations often impact automtic advices
        storedAdviceIndex = self.context.adviceIndex
        self.context._update_after_edit(idxs=[])
        if not self.context.adviceIndex == storedAdviceIndex:
            # we set a status reponse of 500 so the jQuery calling this
            # will refresh the page
            self.request.RESPONSE.status = 500
            return
        return html


class AsyncRenderSearchTerm(BrowserView):
    """ """

    def __call___cachekey(method, self):
        '''cachekey method for self.__call__.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        cfg_modified = cfg.modified()
        # URL to the annex_type can change if server URL changed
        server_url = self.request.get('SERVER_URL', None)
        # cache until a meeting is modified
        date = get_cachekey_volatile('Products.PloneMeeting.Meeting.modified')
        # return meeting.UID if we are on a meeting or None if not
        # as portlet is highlighting the meeting we are on
        meeting_uid = self.context.__class__.__name__ == 'Meeting' and self.context.UID() or None
        collection_uid = self.request.get('collection_uid')
        return (get_plone_groups_for_user(),
                cfg_modified,
                server_url,
                date,
                meeting_uid,
                collection_uid)

    @ram.cache(__call___cachekey)
    def AsyncRenderSearchTerm__call__(self):
        """ """
        self.collection_uid = self.request.get('collection_uid')
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.collection = uuidToObject(self.collection_uid, unrestricted=True)
        self.brains = self.collection.results(batch=False, brains=True)
        rendered_term = ViewPageTemplateFile("templates/term_searchmeetings.pt")(self)
        return rendered_term

    # do ram.cache have a different key name
    __call__ = AsyncRenderSearchTerm__call__


class AsyncLoadLinkedItems(BrowserView):
    """ """

    def __call__(self):
        """ """
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.portal_url = api.portal.get().absolute_url()
        return self.index()


class AsyncLoadLinkedItemsInfos(BrowserView):

    def __call__(self, fieldsConfigAttr, currentCfgId):
        """ """
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.portal_url = api.portal.get().absolute_url()
        # more infos, compute it first to have fields/static fields to show
        more_infos_view = self.context.restrictedTraverse('@@item-more-infos')
        more_infos = more_infos_view(fieldsConfigAttr, currentCfgId)
        # static infos
        static_infos = ''
        visibleColumns = [field for field in more_infos_view.visibleFields
                          if field.startswith('static_')]
        if visibleColumns:
            static_infos = self.context.restrictedTraverse('@@static-infos')(
                visibleColumns=visibleColumns)
        return static_infos + more_infos


class AsyncLoadItemAssemblyAndSignatures(BrowserView):
    """ """

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = api.portal.get()
        self.portal_url = self.portal.absolute_url()
        self._cached_vote_is_secret = {}

    def show(self):
        """ """
        return self.meeting is not None

    def get_cached_vote_is_secret(self, vote_number):
        """ """
        if vote_number not in self._cached_vote_is_secret:
            self._cached_vote_is_secret[vote_number] = self.context.get_vote_is_secret(self.meeting, vote_number)
        return self._cached_vote_is_secret[vote_number]

    def vote_counts(self):
        """Returns informations regarding votes count."""
        data = []
        counts = []
        total_count = 0
        for vote_number in range(len(self.item_votes)):
            sub_counts = []
            total_votes = self.context.get_vote_count(self.meeting, 'any_votable', vote_number)
            number_of_votes_msg = translate(
                'number_of_voters', domain='PloneMeeting', context=self.request)
            res = [u'<span title="{0}">{1}</span>'.format(
                number_of_votes_msg,
                total_votes)]
            formated_total_votes = total_votes
            pattern = u'<span class="vote_value_{0}" title="{1}">{2}</span>'

            # specify how much voted for this vote if secret
            vote_is_secret = self.get_cached_vote_is_secret(vote_number)
            if vote_is_secret:
                voted = self.context.get_vote_count(self.meeting, 'any_voted', vote_number)
                formated_total_votes = "{0} / {1}".format(voted, total_votes)
            sub_counts.append((number_of_votes_msg,
                               formated_total_votes,
                               'vote_value_number_of_voters'))

            # compute votes not encoded for first secret vote
            # taking into account linked votes
            if vote_is_secret:
                linked_vote_numbers = _get_linked_item_vote_numbers(
                    self.context, self.meeting, vote_number) or [0]
                if not linked_vote_numbers or vote_number == min(linked_vote_numbers):
                    total_voted = 0
                    for linked_vote_number in linked_vote_numbers:
                        total_voted += self.context.get_vote_count(self.meeting, 'any_voted', linked_vote_number)
                    translated_not_yet_vote_value = translate(
                        'vote_value_not_yet',
                        domain='PloneMeeting',
                        context=self.request)
                    count = total_votes - total_voted
                    res.append(pattern.format(
                        NOT_ENCODED_VOTE_VALUE,
                        translated_not_yet_vote_value,
                        count))
                    sub_counts.append((translated_not_yet_vote_value,
                                       count,
                                       'vote_value_not_yet'))

            used_vote_terms = get_vocab(
                self.context,
                "Products.PloneMeeting.vocabularies.usedvotevaluesvocabulary",
                vote_number=vote_number)
            for term in used_vote_terms._terms:
                count = self.context.get_vote_count(self.meeting, term.token, vote_number)
                res.append(pattern.format(
                    term.token,
                    term.title,
                    count))
                if term.token != NOT_ENCODED_VOTE_VALUE:
                    total_count += count
                sub_counts.append((term.title, count, 'vote_value_' + term.token))
            votes = u" / ".join(res)
            data.append(votes)
            counts.append(sub_counts)
        return data, counts, total_count

    def compute_next_vote_number(self):
        """Return next vote_number."""
        return len(self.item_votes)

    def show_add_vote_linked_to_previous_icon(self, vote_number):
        """Show add linked_to_previous icon only on last element.
           More over, may only add linked_to_previous if :
           - already a linked_to_previous element;
           - not linked_to_previous element does not use forbidden vote_values,
             aka vote_values not in MeetingConfig.firstLinkedVoteUsedVoteValues;
           - vote poll_type was not redefined."""
        res = False
        vote_infos = self.item_votes[vote_number]
        if (vote_infos['vote_number'] + 1 == self.next_vote_number):
            if vote_infos['linked_to_previous']:
                res = True
            elif vote_infos.get('poll_type') == self.context.getPollType():
                # check vote_values not out of MeetingConfig.firstLinkedVoteUsedVoteValues
                if self.get_cached_vote_is_secret(vote_number):
                    vote_values = [vote_value for vote_value, vote_count
                                   in vote_infos['votes'].items()
                                   if vote_count and vote_value in self.cfg.getUsedVoteValues()]
                else:
                    vote_values = [vote_value for vote_value in vote_infos['voters'].values()]
                if not set(vote_values).difference(
                    self.cfg.getUsedVoteValues(
                        used_values_attr='firstLinkedVoteUsedVoteValues',
                        include_not_encoded=True)):
                    res = True
        return res

    def __call___cachekey(method, self):
        '''cachekey method for self.__call__.
           Cache is invalidated depending on :
           - current user may edit or not;
           - something is redefined for current item or not.'''
        # when using raw fields (assembly, absents, signatures, ...)
        # we invalidate if a raw value changed
        date = get_cachekey_volatile(
            'Products.PloneMeeting.browser.async.AsyncLoadItemAssemblyAndSignaturesRawFields')
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        cfg_modified = cfg.modified()
        # cache will be invalidated if something changed about attendees on meeting
        meeting = self.context.getMeeting()
        ordered_contacts = meeting.ordered_contacts.items()
        redefined_item_attendees = meeting._get_all_redefined_attendees(only_keys=False)
        show_votes = self.context.show_votes()
        context_uid = self.context.UID()
        item_votes_modified = context_uid in meeting.item_votes and meeting.item_votes[context_uid]._p_mtime
        item_attendees_order = meeting._get_item_attendees_order(context_uid)
        # if something redefined for context or not
        if context_uid not in str(redefined_item_attendees):
            redefined_item_attendees = []
        may_change_attendees = self.context._mayChangeAttendees()
        poll_type = self.context.getPollType()
        cache_date = self.request.get('cache_date', None)
        # when using a cache_date, make sure cache is invalidated
        if cache_date:
            date = invalidate_cachekey_volatile_for(
                'Products.PloneMeeting.browser.async.AsyncLoadItemAssemblyAndSignaturesRawFields',
                get_again=True)
        else:
            date = get_cachekey_volatile(
                'Products.PloneMeeting.browser.async.AsyncLoadItemAssemblyAndSignaturesRawFields')
        # when using votesResult, invalidate cache if field content changed
        votesResult = None
        if self.context.attribute_is_used('votesResult'):
            votesResult = self.context.getRawVotesResult(real=True)
        return (date,
                context_uid,
                cfg_modified,
                ordered_contacts,
                redefined_item_attendees,
                item_attendees_order,
                show_votes,
                item_votes_modified,
                may_change_attendees,
                poll_type,
                cache_date,
                votesResult)

    def _update(self):
        """ """
        self.error_msg = self.request.get('attendees_error_msg')
        self.context_uid = self.context.UID()
        self.tool = api.portal.get_tool('portal_plonemeeting')
        # necessary for the @@pm-macros/viewContentField
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.used_item_attrs = self.cfg.getUsedItemAttributes()
        self.member = api.user.get_current()
        self.used_meeting_attrs = self.cfg.getUsedMeetingAttributes()
        self.meeting = self.context.getMeeting()
        self.show_votes = self.context.show_votes()
        if self.show_votes:
            self.voters = self.context.get_item_voters() or []
            self.item_votes = self.context.get_item_votes(
                include_unexisting=True,
                ignored_vote_values=[NOT_VOTABLE_LINKED_TO_VALUE]) or []
            self.voted_voters = self.context.get_voted_voters()
            self.next_vote_number = self.compute_next_vote_number()
            self.displayable_counts, self.counts, self.total_count = self.vote_counts()
        else:
            self.voters = []
            self.item_votes = []
            self.voted_voters = []
            self.next_vote_number = None
            self.counts = None

    @ram.cache(__call___cachekey)
    def AsyncLoadItemAssemblyAndSignatures__call__(self):
        """ """
        self._update()
        return self.index()

    # do ram.cache have a different key name
    __call__ = AsyncLoadItemAssemblyAndSignatures__call__


class AsyncLoadMeetingAssemblyAndSignatures(BrowserView, BaseMeetingView):
    """ """

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = api.portal.get()
        self.portal_url = self.portal.absolute_url()

    def __call___cachekey(method, self):
        '''cachekey method for self.__call__.
           Cache is invalidated depending on :
           - current user may edit or not;
           - something is redefined for current item or not.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        is_manager = tool.isManager(cfg)
        cfg_modified = cfg.modified()
        ordered_contacts = self.context.ordered_contacts.items()
        item_votes = sorted(self.context.get_item_votes().items())
        cache_date = self.request.get('cache_date', None)
        # when using a cache_date, make sure cache is invalidated
        if cache_date:
            date = invalidate_cachekey_volatile_for(
                'Products.PloneMeeting.browser.async.AsyncLoadMeetingAssemblyAndSignatures',
                get_again=True)
        else:
            date = get_cachekey_volatile(
                'Products.PloneMeeting.browser.async.AsyncLoadMeetingAssemblyAndSignatures')
        return (date,
                is_manager,
                cfg_modified,
                ordered_contacts,
                item_votes,
                repr(self.context),
                cache_date)

    def _update(self):
        """ """
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        self.used_attrs = self.cfg.getUsedMeetingAttributes()
        self.show_voters = self.cfg.getUseVotes()
        self.voters = self.context.get_voters()
        view = self.context.restrictedTraverse('@@view')
        view.update()
        self.meeting_view = view

    @ram.cache(__call___cachekey)
    def AsyncLoadMeetingAssemblyAndSignatures__call__(self):
        """ """
        self._update()
        return self.index()

    # do ram.cache have a different key name
    __call__ = AsyncLoadMeetingAssemblyAndSignatures__call__
